from __future__ import annotations

import typing

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.http import Http404

from restdoctor.rest_framework.generics import GenericAPIView
from restdoctor.rest_framework.mixins import ListModelMixin, RetrieveModelMixin
from restdoctor.rest_framework.sensitive_data import clear_sensitive_data
from restdoctor.rest_framework.signals import bind_extra_request_view_initial_metadata
from restdoctor.utils.permissions import get_permission_classes_from_map
from restdoctor.utils.sentry import capture_exception
from restdoctor.utils.serializers import get_serializer_class_from_map
from restdoctor.utils.structlog import get_logger

if typing.TYPE_CHECKING:
    from django.http import HttpRequest
    from django.core.handlers.wsgi import WSGIRequest
    from rest_framework.permissions import BasePermission
    from rest_framework.response import Response
    from rest_framework.request import Request
    from rest_framework.serializers import BaseSerializer
    from restdoctor.utils.serializers import SerializerType
    from restdoctor.rest_framework.sensitive_data import SerializerData


logger = get_logger(__name__)


class SerializerClassMapApiView(GenericAPIView):
    action_map: typing.Dict[str, str] = {}
    action: str = ''
    permission_classes_map: typing.Dict[str, typing.List[BasePermission]]

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        if 'permission_classes' in kwargs and getattr(self, 'permission_classes_map', None):
            raise ImproperlyConfigured('Use permission_classes_map')
        super().__init__(*args, **kwargs)

    @classmethod
    def as_view(cls, **initkwargs: typing.Any) -> SerializerClassMapApiView:
        view = super().as_view(**initkwargs)
        if cls.action_map:
            view.actions = cls.action_map
        return view

    def dispatch(self, request: WSGIRequest, *args: typing.Any, **kwargs: typing.Any) -> Response:
        response = super().dispatch(request, *args, **kwargs)
        response.serializer = self.get_response_serializer_class()
        return response

    def clear_request_data(self, request: Request) -> typing.Optional[SerializerData]:
        request_serializer = self.get_request_serializer_class()
        request_data = request.data
        if request_data and request_serializer:
            return clear_sensitive_data(request_data, request_serializer)
        return None

    def initial(self, request: Request, *args: typing.Any, **kwargs: typing.Any) -> None:
        if settings.API_ENABLE_STRUCTLOG:
            bind_extra_request_view_initial_metadata.send(sender=self.__class__, request=request, logger=logger)
            try:
                request_data = self.clear_request_data(request)
            except Exception:
                capture_exception()
                request_data = None

            logger.info(
                'view_initial',
                app_name=self.__module__.split('.')[0],
                module=self.__module__,
                view_name=self.__class__.__name__,
                request_data=request_data,
                request_query_params=dict(request.query_params),
                action=self.get_action(),
            )
        super().initial(request, *args, **kwargs)

    def get_action(self) -> str:
        action = getattr(self, 'action', None)
        if action:
            return action
        if self.request:
            return self.request.method.lower()
        return 'default'

    def get_permissions(self) -> typing.List[BasePermission]:
        permission_classes = get_permission_classes_from_map(
            action=self.get_action(),
            permission_classes_map=getattr(self, 'permission_classes_map', {}),
            default_permission_classes=self.permission_classes,
        )
        return [permission() for permission in permission_classes]

    def get_serializer_class(
        self, stage: str = 'response', action: str = None, api_format: str = None, use_default: bool = True,
    ) -> SerializerType:
        action = action or self.get_action()
        serializer_class_map = getattr(self, 'serializer_class_map', {})

        if api_format is None:
            api_format = settings.API_DEFAULT_FORMAT
            if self.request and self.request.api_params:
                api_format = self.request.api_params.format or api_format

        return get_serializer_class_from_map(
            action, stage, serializer_class_map, self.serializer_class, use_default=use_default, api_format=api_format)

    def get_serializer_context(self, stage: str = 'response') -> typing.Dict[str, typing.Any]:
        return super().get_serializer_context()

    def get_serializer_instance(
        self, serializer_class: SerializerType,
        *args: typing.Any, **kwargs: typing.Any,
    ) -> BaseSerializer:
        stage = kwargs.pop('stage', 'response')
        kwargs['context'] = self.get_serializer_context(stage)
        return serializer_class(*args, **kwargs)

    def get_request_serializer_class(self, use_default: bool = True) -> SerializerType:
        return self.get_serializer_class('request', use_default=use_default)

    def get_response_serializer_class(self) -> SerializerType:
        return self.get_serializer_class('response')

    def get_request_serializer(
            self, *args: typing.Any, use_default: bool = True, **kwargs: typing.Any,
    ) -> BaseSerializer:
        stage = 'request'
        serializer_class = self.get_serializer_class(stage, use_default=use_default)
        return self.get_serializer_instance(serializer_class, stage=stage, *args, **kwargs)

    def get_response_serializer(self, *args: typing.Any, **kwargs: typing.Any) -> BaseSerializer:
        stage = 'response'
        serializer_class = self.get_serializer_class(stage)
        return self.get_serializer_instance(serializer_class, stage=stage, *args, **kwargs)


class ListAPIView(ListModelMixin, SerializerClassMapApiView):
    action_map = {'get': 'list'}

    def get(self, request: HttpRequest, *args: typing.Any, **kwargs: typing.Any) -> Response:
        self.action = 'list'
        return self.list(request, *args, **kwargs)


class RetrieveAPIView(RetrieveModelMixin, SerializerClassMapApiView):
    action_map = {'get': 'retrieve'}

    def get(self, request: HttpRequest, *args: typing.Any, **kwargs: typing.Any) -> Response:
        self.action = 'retrieve'
        return self.retrieve(request, *args, **kwargs)


class NotFound(GenericAPIView):
    def get(self, request: HttpRequest, *args: typing.Any, **kwargs: typing.Any) -> Response:
        raise Http404(f'Resource {request.path} not found')



from rest_framework.renderers import JSONRenderer

from restdoctor.rest_framework.response import ResponseWithMeta


def test_response_with_meta_getstate():
    response = ResponseWithMeta(data={}, meta={})
    response.renderer_context = {}
    response.accepted_renderer = JSONRenderer()
    response.accepted_media_type = 'application/json'
    response.render()

    state = response.__getstate__()

    assert 'meta' not in state
    assert 'data' in state



from __future__ import annotations
import logging
import typing

logger = logging.getLogger(__name__)

if typing.TYPE_CHECKING:
    from restdoctor.rest_framework.schema.custom_types import LocalRefs, OpenAPISchema


class LocalRefsRegistry:
    def __init__(self) -> None:
        self._local_refs: LocalRefs = {}

    def put_local_ref(self, ref: str, schema: OpenAPISchema) -> None:
        if ref.startswith('#/components'):
            path = tuple(ref.split('/')[2:])
            if path[-1] == '':
                logger.error(f'Wrong ref {ref} {schema}')
            if path in self._local_refs:
                if self._local_refs[path] != schema:
                    logger.error(f'Replacing schema for {path}')
            else:
                self._local_refs[path] = schema

    def get_components(self) -> OpenAPISchema:
        components: OpenAPISchema = {}
        for path, schema in self._local_refs.items():
            component = components
            for path_item in path[:-1]:
                try:
                    component = component[path_item]
                except KeyError:
                    components[path_item] = {}
                    component = component[path_item]
            component[path[-1]] = schema
        return components



from __future__ import annotations

import typing

from rest_framework.fields import Field


class SchemaWrapper(Field):
    def __new__(
        cls, wrapped: Field, schema_type: typing.Union[Field, typing.Type[Field]] = None,
    ) -> Field:
        if isinstance(schema_type, type):
            schema_type = schema_type()
        wrapped.schema_type = schema_type
        return wrapped



from restdoctor.rest_framework.routers import ResourceRouter
from tests.test_unit.test_schema.stubs import DefaultViewSet, ViewSetWithTags


def test_schema_with_tags_success_case(get_create_view_func):
    create_view = get_create_view_func('test', ViewSetWithTags, 'test')

    view = create_view('/test/', 'GET')
    operation = view.schema.get_operation('/test/', 'GET')

    assert operation['tags'] == ['tag1', 'tag2']


def test_schema_tags_from_path_success_case(get_create_view_func):
    create_view = get_create_view_func('test', DefaultViewSet, 'test')

    view = create_view('/test/', 'GET')
    operation = view.schema.get_operation('/test/', 'GET')

    assert operation['tags'] == ['test']


def test_resource_get_tags_from_default_success_case(get_create_view_func, resource_class, settings):
    settings.API_RESOURCE_DEFAULT = 'default'
    viewset_class = resource_class('default', ViewSetWithTags)
    create_view = get_create_view_func('test', viewset_class, 'test', router=ResourceRouter())

    view = create_view('/test/', 'GET')
    operation = view.schema.get_operation('/test/', 'GET')

    assert operation['tags'] == ['tag1', 'tag2']



from rest_framework.fields import CharField
from restdoctor.rest_framework.serializers import ModelSerializer, Serializer
from tests.stubs.models import MyModel


class ParentModelSerializer(ModelSerializer):
    class Meta:
        model = MyModel
        fields = ['field_parent_1', 'field_parent_2']

    field_parent_1 = CharField()
    field_parent_2 = CharField()


class ParentSerializer(Serializer):
    class Meta:
        fields = ['field_parent_1', 'field_parent_2']

    field_parent_1 = CharField()
    field_parent_2 = CharField()


class SerializerMixinA(Serializer):
    class Meta:
        fields = ['field_mixin_a']

    field_mixin_a = CharField()
    field_mixin_a_not_in_meta = CharField()


class SerializerMixinB(Serializer):
    field_mixin_b = CharField()



from django.apps import AppConfig as BaseConfig
from django.conf import settings

from restdoctor.rest_framework.fields import DateTimeField
from restdoctor import app_settings


class AppConfig(BaseConfig):
    name = 'restdoctor'

    def ready(self) -> None:
        from django.db import models
        from rest_framework.serializers import ModelSerializer

        field_mapping = ModelSerializer.serializer_field_mapping
        field_mapping[models.DateTimeField] = DateTimeField

        for setting in dir(app_settings):
            if setting.isupper() and not hasattr(settings, setting):
                setattr(settings, setting, getattr(app_settings, setting))



import os

import django
from django.test.utils import setup_databases


def pytest_configure(config):
    os.environ['DJANGO_SETTINGS_MODULE'] = 'tests.test_settings'
    django.setup()
    setup_databases(verbosity=1, interactive=False)



from typing import List

from django.db.models import Model


def get_model_sensitive_fields(model: Model) -> List[str]:
    try:
        return model.SensitiveData.include
    except AttributeError:
        return []


def is_model_field_sensitive(model: Model, field_path: str) -> bool:
    field_name, *related_field_name = field_path.split('.', 1)
    if related_field_name:
        try:
            return is_model_field_sensitive(
                getattr(model, field_name).field.related_model, related_field_name[0],
            )
        except AttributeError:
            return False
    else:
        return field_name in get_model_sensitive_fields(model)



from django.conf import settings

API_FALLBACK_VERSION = 'fallback'
API_FALLBACK_FOR_APPLICATION_JSON_ONLY = False
API_DEFAULT_VERSION = 'default'
API_DEFAULT_FORMAT = 'full'
API_PREFIXES = ('/api',)
API_FORMATS = ('full',)

API_RESOURCE_DISCRIMINATIVE_PARAM = 'view_type'
API_RESOURCE_DEFAULT = 'default'
API_RESOURCE_SET_PARAM = False
API_RESOURCE_SET_PARAM_FOR_DEFAULT = False

API_VERSIONS = {
    API_FALLBACK_VERSION: settings.ROOT_URLCONF,
    API_DEFAULT_VERSION: settings.ROOT_URLCONF,
}

API_ENABLE_STRUCTLOG = getattr(settings, 'ENABLE_STRUCTLOG', False)

API_STRICT_SCHEMA_VALIDATION = getattr(settings, 'API_STRICT_SCHEMA_VALIDATION', False)



import pytest

from restdoctor.utils.media_type import parse_accept


def test_parse_none():
    result = parse_accept(None)

    assert result is None


@pytest.mark.parametrize(
    'header,expected_version',
    (
        ('*/*', 'fallback'),
        ('text/html', 'fallback'),
        ('application/json', 'fallback'),
        ('application/vnd.vendor', 'default'),
    ),
)
def test_parse_accept_fallback(settings, header, expected_version):
    settings.API_FALLBACK_VERSION = 'fallback'
    settings.API_DEFAULT_VERSION = 'default'
    settings.API_VERSIONS = {
        'v1': 'v1',
        'v2': 'v2',
    }
    settings.API_FALLBACK_FOR_APPLICATION_JSON_ONLY = False

    result = parse_accept(header)

    assert expected_version == result.version


@pytest.mark.parametrize(
    'header,expected_version',
    (
        ('*/*', 'default'),
        ('text/html', 'default'),
        ('application/json', 'fallback'),
        ('application/vnd.vendor', 'default'),
    ),
)
def test_parse_accept_default(settings, header, expected_version):
    settings.API_FALLBACK_VERSION = 'fallback'
    settings.API_DEFAULT_VERSION = 'default'
    settings.API_VERSIONS = {
        'v1': 'v1',
        'v2': 'v2',
    }
    settings.API_FALLBACK_FOR_APPLICATION_JSON_ONLY = True

    result = parse_accept(header)

    assert expected_version == result.version


@pytest.mark.parametrize(
    ('header', 'expected_version', 'expected_discriminator'),
    [
        ('application/vnd.vendor', 'default', None),
        ('application/vnd.vendor.v1', 'v1', None),
        ('application/vnd.vendor.v1.full', 'v1', None),
        ('application/vnd.vendor.v1.full+json', 'v1', None),
        ('application/vnd.vendor.v1-extended', 'v1', 'extended'),
        ('application/vnd.vendor.v1-extended.full', 'v1', 'extended'),
        ('application/vnd.vendor.v1-extended.full+json', 'v1', 'extended'),
        ('application/vnd.vendor.v2', 'v2', None),
        ('application/vnd.vendor.v2.full', 'v2', None),
        ('application/vnd.vendor.v2.full+json', 'v2', None),
        ('application/vnd.vendor.v2-extended', 'v2', 'extended'),
        ('application/vnd.vendor.v2-extended.full', 'v2', 'extended'),
        ('application/vnd.vendor.v2-extended.full+json', 'v2', 'extended'),
        ('application/vnd.vendor.v3', 'default', None),
        ('application/vnd.vendor.v3.full', 'default', None),
        ('application/vnd.vendor.v3.full+json', 'default', None),
        ('application/vnd.vendor.v3-extended', 'default', 'extended'),
        ('application/vnd.vendor.v3-extended.full', 'default', 'extended'),
        ('application/vnd.vendor.v3-extended.full+json', 'default', 'extended'),
    ],
)
def test_parse_accept_versions_and_resource_discriminator(settings, header, expected_version, expected_discriminator):
    settings.API_FALLBACK_VERSION = 'fallback'
    settings.API_DEFAULT_VERSION = 'default'
    settings.API_RESOURCE_DEFAULT = 'common'
    settings.API_VERSIONS = {
        'v1': 'v1',
        'v2': 'v2',
    }

    result = parse_accept(header)

    assert expected_version == result.version
    assert expected_discriminator == result.resource_discriminator


@pytest.mark.parametrize(
    'vendor,expected_vendor',
    (
        (None, 'vendor'),
        ('', 'vendor'),
        ('my_vendor', 'my_vendor'),
    ),
)
def test_parse_accept_vendor_success_case(vendor, expected_vendor):
    result = parse_accept('application/vnd', vendor=vendor)

    assert expected_vendor == result.vendor



from __future__ import annotations
import typing

from django.core.exceptions import PermissionDenied
from django.http import Http404
from rest_framework import exceptions
from rest_framework.response import Response
from rest_framework.views import set_rollback

from restdoctor.utils.helpers import get_full_class_name

if typing.TYPE_CHECKING:
    from restdoctor.utils.custom_types import ImmutableContext, GenericContext


def _get_full_details(detail: typing.Any) -> typing.Union[typing.List, typing.Dict]:  # noqa: CCR001
    if isinstance(detail, list):
        return [_get_full_details(item) for item in detail]
    elif isinstance(detail, dict):
        for field, value in detail.items():
            details = _get_full_details(value)
            for item in details:
                if not isinstance(item, dict):
                    continue
                item['field'] = field
            return details
    return {
        'message': detail,
        'code': detail.code,
    }


def _get_errors_data(exc: exceptions.APIException) -> GenericContext:
    full_details = _get_full_details(exc.detail)
    errors_data = {'message': exc.default_detail}
    if isinstance(full_details, list):
        errors_data['errors'] = full_details
    elif isinstance(full_details, dict):
        errors_data['errors'] = [full_details]
    return errors_data


def _override_exception(exc: Exception) -> Exception:
    if isinstance(exc, Http404):
        exc = exceptions.NotFound(exc)
    elif get_full_class_name(exc) == 'ratelimit.exceptions.Ratelimited':
        exc = exceptions.PermissionDenied(
            detail={
                'message': 'Вы превысили лимит попыток. Попробуйте позже или обратитесь в поддержку.',
            },
            code='too_many_requests',
        )
    elif isinstance(exc, PermissionDenied):
        exc = exceptions.PermissionDenied()
    return exc


def exception_handler(exc: Exception, context: ImmutableContext) -> typing.Optional[Response]:
    exc = _override_exception(exc)

    if isinstance(exc, exceptions.APIException):
        response_data = _get_errors_data(exc)

        headers = {}
        if getattr(exc, 'auth_header', None):
            headers['WWW-Authenticate'] = exc.auth_header

        wait = getattr(exc, 'wait', None)
        if wait:
            headers['Retry-After'] = wait
            response_data['retry_after'] = wait

        set_rollback()
        return Response(
            response_data,
            status=exc.status_code,
            headers=headers,
        )

    return None



import pytest

from restdoctor.rest_framework.serializers import EmptySerializer
from tests.test_unit.stubs import (
    ListViewSetWithRequestSerializer, SerializerB, ListViewSetWithoutRequestSerializer,
    SerializerA,
)


@pytest.mark.parametrize(
    'viewset,use_default,expected',
    (
            (ListViewSetWithRequestSerializer, True, SerializerB),
            (ListViewSetWithRequestSerializer, False, SerializerB),
            (ListViewSetWithoutRequestSerializer, True, SerializerA),
            (ListViewSetWithoutRequestSerializer, False, EmptySerializer),
    ),
)
def test_get_request_serializer_class(viewset, use_default, expected):
    list_view = viewset(request=None, action='list')

    request_serializer_class = list_view.get_request_serializer_class(use_default=use_default)

    assert request_serializer_class == expected


@pytest.mark.parametrize(
    'viewset,use_default,expected',
    (
            (ListViewSetWithRequestSerializer, True, SerializerB),
            (ListViewSetWithRequestSerializer, False, SerializerB),
            (ListViewSetWithoutRequestSerializer, True, SerializerA),
            (ListViewSetWithoutRequestSerializer, False, EmptySerializer),
    ),
)
def test_get_request_serializer(viewset, use_default, expected):
    list_view = viewset(request=None, action='list', format_kwarg=None)

    request_serializer = list_view.get_request_serializer(use_default=use_default)

    assert isinstance(request_serializer, expected)



from __future__ import annotations
import typing

from rest_framework.response import Response


class ResponseWithMeta(Response):
    def __init__(
        self, data: typing.Any, meta: typing.Dict[str, typing.Any] = None,
        *args: typing.Any, **kwargs: typing.Any,
    ) -> None:
        self.meta = meta
        super().__init__(data, *args, **kwargs)

    @property
    def rendered_content(self) -> bytes:
        context = getattr(self, 'renderer_context', None)
        if context:
            context['meta'] = self.meta

        return super().rendered_content

    def __getstate__(self) -> typing.Dict[str, typing.Any]:
        state = super().__getstate__()
        for key in (
            'meta',
        ):
            if key in state:
                del state[key]
        return state



from __future__ import annotations

import typing

from rest_framework import status
from rest_framework.mixins import (
    ListModelMixin as BaseListModelMixin,
    RetrieveModelMixin as BaseRetrieveModelMixin,
    CreateModelMixin as BaseCreateModelMixin,
    UpdateModelMixin as BaseUpdateModelMixin,
    DestroyModelMixin as BaseDestroyModelMixin,
)
from rest_framework.request import Request
from rest_framework.response import Response

from restdoctor.rest_framework.negotiations import APIVersionContentNegotiation
from restdoctor.rest_framework.pagination import PageNumberPagination

if typing.TYPE_CHECKING:
    from django.db.models import QuerySet
    from rest_framework.pagination import BasePagination
    from rest_framework.serializers import BaseSerializer


class NegotiatedMixin:
    content_negotiation_class = APIVersionContentNegotiation


class CreateModelMixin(BaseCreateModelMixin):
    def create(self, request: Request, *args: typing.Any, **kwargs: typing.Any) -> Response:
        request_serializer_class = self.get_request_serializer_class()
        serializer = self.get_serializer_instance(
            request_serializer_class, stage='request', data=request.data)
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)

        response_serializer_class = self.get_response_serializer_class()
        if response_serializer_class != request_serializer_class:
            serializer = self.get_serializer_instance(
                response_serializer_class, serializer.instance, stage='response')
        headers = self.get_success_headers(serializer.data)
        return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)


class ListModelMixin(BaseListModelMixin):
    pagination_class: typing.Optional[BasePagination] = PageNumberPagination

    def get_serializer(self, *args: typing.Any, **kwargs: typing.Any) -> BaseSerializer:
        return self.get_response_serializer(*args, **kwargs)

    def list(self, request: Request, *args: typing.Any, **kwargs: typing.Any) -> Response:  # noqa: A003
        request_serializer = self.get_request_serializer(data=request.query_params, use_default=False)
        request_serializer.is_valid(raise_exception=True)
        queryset = self.get_collection(request_serializer)

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(queryset, many=True)
        return Response(serializer.data)

    def get_collection(self, request_serializer: BaseSerializer) -> typing.Union[typing.List, QuerySet]:
        return self.filter_queryset(self.get_queryset())


class RetrieveModelMixin(BaseRetrieveModelMixin):
    def get_serializer(self, *args: typing.Any, **kwargs: typing.Any) -> BaseSerializer:
        return self.get_response_serializer(*args, **kwargs)


class UpdateModelMixin(BaseUpdateModelMixin):
    def update(self, request: Request, *args: typing.Any, **kwargs: typing.Any) -> Response:
        partial = kwargs.pop('partial', False)
        instance = self.get_object()
        request_serializer_class = self.get_request_serializer_class()
        serializer = self.get_serializer_instance(
            request_serializer_class, instance, stage='request', data=request.data, partial=partial)
        serializer.is_valid(raise_exception=True)
        self.perform_update(serializer)

        if getattr(serializer.instance, '_prefetched_objects_cache', None):
            serializer.instance._prefetched_objects_cache = {}

        response_serializer_class = self.get_response_serializer_class()
        if response_serializer_class != request_serializer_class:
            serializer = self.get_serializer_instance(
                response_serializer_class, serializer.instance, stage='response')
        return Response(serializer.data)


class DestroyModelMixin(BaseDestroyModelMixin):
    pass



from __future__ import annotations
import typing

from django.conf import settings

from restdoctor.utils.api_prefix import get_api_prefixes
from restdoctor.utils.media_type import parse_accept, get_api_header

if typing.TYPE_CHECKING:
    from django.http import HttpResponse, HttpRequest
    from restdoctor.django.custom_types import DjangoHandler


class ApiSelectorMiddleware:
    def __init__(self, get_response: DjangoHandler):
        self.api_versions = settings.API_VERSIONS
        self.api_fallback_version = settings.API_FALLBACK_VERSION
        self.fallback_urlconf = self.api_versions.get(self.api_fallback_version, settings.ROOT_URLCONF)

        self.api_vendor_string = getattr(settings, 'API_VENDOR_STRING', 'Vendor')
        self.api_vendor_accept = self.api_vendor_string.lower()

        self.get_response = get_response
        self.api_prefixes = get_api_prefixes(default=None)
        self.schema_prefixes = tuple(
            f'{api_prefix}/openapi.schema' for api_prefix in (self.api_prefixes or [])
        )

    def is_api_call(self, request: HttpRequest) -> bool:
        if self.api_prefixes:
            return request.path_info.startswith(self.api_prefixes)

    def is_schema_call(self, request: HttpRequest) -> bool:
        return request.path_info.startswith(self.schema_prefixes)

    def __call__(self, request: HttpRequest) -> HttpResponse:
        if not self.is_api_call(request):
            return self.get_response(request)

        if self.is_schema_call(request):
            api_params = parse_accept(
                f'application/vnd.{self.api_vendor_accept}', vendor=self.api_vendor_accept)
        else:
            api_params = parse_accept(
                request.headers.get('accept'), vendor=self.api_vendor_accept)
        request.api_params = api_params
        api_version = (api_params and api_params.version) or self.api_fallback_version
        request.urlconf = self.api_versions.get(api_version, self.fallback_urlconf)

        response = self.get_response(request)
        if api_params is not None:
            response[f'X-{self.api_vendor_string}-Media-Type'] = get_api_header(api_params)

        return response



from __future__ import annotations
import typing

from django.utils.functional import lazy
from rest_framework.reverse import reverse as drf_reverse
from rest_framework.utils.urls import replace_query_param

if typing.TYPE_CHECKING:
    from rest_framework.request import Request
    from restdoctor.utils.custom_types import GenericContext


def preserve_resource_params(url: str, request: Request) -> str:
    if not request:
        return url

    args = getattr(request, 'resource_args', {})
    for param, value in args.items():
        url = replace_query_param(url, param, value)

    return url


def reverse(
    viewname: str,
    args: typing.List[typing.Any] = None, kwargs: GenericContext = None, request: Request = None,
    **extra: typing.Any,
) -> str:
    url = drf_reverse(viewname, args=args, kwargs=kwargs, request=request, **extra)
    return preserve_resource_params(url, request)


reverse_lazy = lazy(reverse, str)



import pytest


from restdoctor.django.sensitive_data import get_model_sensitive_fields, is_model_field_sensitive
from restdoctor.rest_framework.sensitive_data import (
    get_serializer_sensitive_fields, get_serializer_instance, get_serializer_sensitive_data_config,
    get_field_serializer, clear_sensitive_data,
)
from tests.test_unit.stubs import (
    ParentSensitiveDataModel, ModelWithoutSensitiveData,
    SerializerWithSensitiveData, SerializerWithoutSensitiveData, ModelSerializerWithSensitiveData,
)

@pytest.mark.parametrize(
    ('model', 'expected_result'),
    (
        (ParentSensitiveDataModel, ['title']),
        (ModelWithoutSensitiveData, []),
    ),
)
def test_get_model_sensitive_fields(model, expected_result):
    result = get_model_sensitive_fields(model)

    assert result == expected_result


@pytest.mark.parametrize(
    ('model', 'field_path', 'expected_result'),
    (
        (ParentSensitiveDataModel, 'title.test.test', False),
        (ParentSensitiveDataModel, 'test', False),
        (ParentSensitiveDataModel, 'field_fk', False),
        (ParentSensitiveDataModel, 'title', True),
        (ParentSensitiveDataModel, 'field_fk.title', True),
        (ParentSensitiveDataModel, 'field_fk.field_fk.title', False),
        (ParentSensitiveDataModel, 'field_m2m.title', True),
        (ParentSensitiveDataModel, 'field_o2o.title', True),
    ),
)
def test_is_model_field_sensitive(model, field_path, expected_result):
    result = is_model_field_sensitive(model, field_path)

    assert result == expected_result


@pytest.mark.parametrize(
    ('serializer', 'with_model_sensitive_fields', 'expected_result'),
    (
        (SerializerWithoutSensitiveData, True, []),
        (SerializerWithoutSensitiveData, False, []),
        (SerializerWithSensitiveData, True, ['field1', 'field2']),
        (SerializerWithSensitiveData, False, ['field1', 'field2']),
        (ModelSerializerWithSensitiveData, True, ['first_name', 'last_name', 'title']),
        (ModelSerializerWithSensitiveData, False, ['first_name', 'last_name']),
    ),
)
def test_get_serializer_sensitive_fields(serializer, with_model_sensitive_fields, expected_result):
    result = get_serializer_sensitive_fields(
        serializer=serializer, with_model_sensitive_fields=with_model_sensitive_fields)

    assert result == expected_result


@pytest.mark.parametrize(
    'serializer',
    (SerializerWithSensitiveData, SerializerWithSensitiveData()),
)
def test_get_serializer_instance(serializer):
    result = get_serializer_instance(serializer)

    assert isinstance(result, SerializerWithSensitiveData)


@pytest.mark.parametrize(
    ('field_name', 'field'),
    (
        ('field1', ModelSerializerWithSensitiveData().fields['field1']),
        ('field2', ModelSerializerWithSensitiveData().fields['field2']),
        ('field3', ModelSerializerWithSensitiveData().fields['field3']),
    ),
)
def test_get_field_serializer(field_name, field):
    result = get_field_serializer(ModelSerializerWithSensitiveData(), field_name, field)

    assert isinstance(result, SerializerWithSensitiveData)


def test_get_serializer_sensitive_data_config():
    result = get_serializer_sensitive_data_config(ModelSerializerWithSensitiveData())

    assert result == {
        'first_name': True,
        'last_name': True,
        'title': True,
        'field1': {'field1': True, 'field2': True},
        'field2': {'field1': True, 'field2': True},
        'field3': {'field1': True, 'field2': True},
    }


@pytest.mark.parametrize(
    'data, expected_data',
    (
        ({'first_name': 'Иван', 'id': 1}, {'first_name': '[Cleaned]', 'id': 1}),
        ({'last_name': 'Иванович', 'id': 1}, {'last_name': '[Cleaned]', 'id': 1}),
        (
            [{'first_name': 'Иван', 'id': 1}, {'first_name': 'Петр', 'id': 2}],
            [{'first_name': '[Cleaned]', 'id': 1}, {'first_name': '[Cleaned]', 'id': 2}],
        ),
        (
            {'field1': {'field1': 'Иван', 'id': 1}, 'id': 1},
            {'field1': {'field1': '[Cleaned]', 'id': 1}, 'id': 1},
        ),
    ),
)
def test_clear_sensitive_data(data, expected_data):
    result = clear_sensitive_data(data, ModelSerializerWithSensitiveData())

    assert result == expected_data



from __future__ import annotations
import typing

if typing.TYPE_CHECKING:
    from restdoctor.rest_framework.schema.utils import ActionCodesMap


# True и False - относится ли action к коллекции или отдельному элементу
ACTIONS_MAP = {
    True: {'get': 'list', 'post': 'create'},
    False: {'get': 'retrieve', 'put': 'update', 'patch': 'partial_update', 'delete': 'destroy'},
}

ACTION_CODES_MAP: ActionCodesMap = {
    'default': {
        '400': ('Ошибка валидации запроса.', {'$ref': '#/components/schemas/ErrorResponseSchema'}),
        '404': ('Ресурс не найден.', {'$ref': '#/components/schemas/NotFoundResponseSchema'}),
    },
    'get': {
        '200': 'Успешный запрос.'},
    'post': {
        '200': 'Успешный запрос.'},
    'patch': {
        '200': 'Успешный запрос.'},
    'put': {
        '200': 'Успешный запрос.'},
    'delete': {
        '204': 'Успешный запрос.'},
    'list': {
        '200': 'Успешный запрос коллекции.',
        '400': None,
        '404': None,
    },
    'retrieve': {
        '200': 'Успешный запрос объекта.'},
    'update': {
        '200': 'Успешное изменение объекта.'},
    'partial_update': {
        '200': 'Успешное изменение объекта.'},
    'create': {
        '201': 'Успешное создание объекта.'},
    'destroy': {
        '204': ('Успешное удаление объекта.', {}),
        '400': None,
    },
}



import pytest
from rest_framework.exceptions import NotFound
from rest_framework.request import Request

from restdoctor.constants import DEFAULT_MAX_PAGE_SIZE, DEFAULT_PAGE_SIZE


@pytest.mark.parametrize(
    'page,per_page,expected_size,expected_has_next,expected_has_prev',
    (
        (1, 0, 0, True, False),
        (1, 10, 10, True, False),
        (2, 10, 5, False, True),
        (2, 5, 5, True, True),
        (1, 200, 15, False, False),
        ('', 200, 15, False, False),
    ),
)
@pytest.mark.django_db
def test_page_number_pagination_paginate_queryset(
    mocker, n_models, rf, page_number_pagination, my_models_queryset, page, per_page,
    expected_size, expected_has_next, expected_has_prev,
):
    n_models(15)
    mocked_count = mocker.patch.object(my_models_queryset, 'count')
    mocked_count.return_value = 15
    request = Request(rf.get('/endpoint', data={'page': page, 'per_page': per_page}))

    paginated = page_number_pagination.paginate_queryset(my_models_queryset, request)
    paginated_response = page_number_pagination.get_paginated_response([])
    meta = paginated_response.meta

    assert len(paginated) == expected_size
    assert page_number_pagination.has_next == expected_has_next
    assert page_number_pagination.has_prev == expected_has_prev
    assert (meta['next_url'] is not None) == expected_has_next
    assert (meta['prev_url'] is not None) == expected_has_prev
    mocked_count.assert_called()


@pytest.mark.parametrize(
    'page,per_page,expected_size,expected_has_next,expected_has_prev',
    (
        (1, 10, 10, True, False),
        (2, 10, 5, False, True),
        (2, 5, 5, True, True),
        (1, 200, 15, False, False),
        ('', 200, 15, False, False),
    ),
)
@pytest.mark.django_db
def test_page_number_uncounted_pagination_paginate_queryset(
    mocker, n_models, rf, page_number_uncounted_pagination, my_models_queryset, page, per_page,
    expected_size, expected_has_next, expected_has_prev,
):
    n_models(15)
    mocked_count = mocker.patch.object(my_models_queryset, 'count')
    request = Request(rf.get('/endpoint', data={'page': page, 'per_page': per_page}))

    paginated = page_number_uncounted_pagination.paginate_queryset(my_models_queryset, request)

    assert len(paginated) == expected_size
    assert page_number_uncounted_pagination.has_next == expected_has_next
    assert page_number_uncounted_pagination.has_prev == expected_has_prev
    mocked_count.assert_not_called()


@pytest.mark.parametrize(
    'per_page,expected_per_page',
    (
        (10, 10),
        (100500, DEFAULT_MAX_PAGE_SIZE),
    ),
)
@pytest.mark.django_db
def test_max_per_page(rf, page_number_pagination, my_models_queryset, per_page, expected_per_page):
    request = Request(rf.get('/endpoint', data={'per_page': per_page}))

    _ = page_number_pagination.paginate_queryset(my_models_queryset, request)

    assert page_number_pagination.per_page == expected_per_page


@pytest.mark.django_db
def test_default_per_page(rf, page_number_pagination, my_models_queryset):
    request = Request(rf.get('/endpoint'))

    _ = page_number_pagination.paginate_queryset(my_models_queryset, request)

    assert page_number_pagination.per_page == DEFAULT_PAGE_SIZE


@pytest.mark.parametrize(
    'per_page',
    (
        ('',),
        ('None',),
    ),
)
@pytest.mark.django_db
def test_null_per_page(rf, page_number_pagination, my_models_queryset, per_page):
    request = Request(rf.get('/endpoint', data={'per_page': per_page}))

    _ = page_number_pagination.paginate_queryset(my_models_queryset, request)

    assert page_number_pagination.per_page == DEFAULT_PAGE_SIZE


@pytest.mark.parametrize(
    'page,expected_page',
    (
        (-1, 1),
        (0, 1),
        (1, 1),
        (2, 2),
    ),
)
@pytest.mark.django_db
def test_page_succes_case(rf, n_models, page_number_pagination, my_models_queryset, page, expected_page):
    n_models(15)
    request = Request(rf.get('/endpoint', data={'page': page, 'per_page': 10}))

    _ = page_number_pagination.paginate_queryset(my_models_queryset, request)

    assert page_number_pagination.page == expected_page


@pytest.mark.django_db
def test_incorrect_page_fail_case(rf, n_models, page_number_pagination, my_models_queryset):
    n_models(15)
    request = Request(rf.get('/endpoint', data={'page': 3, 'per_page': 10}))

    with pytest.raises(NotFound):
        _ = page_number_pagination.paginate_queryset(my_models_queryset, request)


@pytest.mark.django_db
def test_default_page(rf, page_number_pagination, my_models_queryset):
    request = Request(rf.get('/endpoint'))

    _ = page_number_pagination.paginate_queryset(my_models_queryset, request)

    assert page_number_pagination.page == 1


@pytest.mark.parametrize(
    'page',
    (
        ('',),
        ('None',),
    ),
)
@pytest.mark.django_db
def test_null_page(rf, page_number_pagination, my_models_queryset, page):
    request = Request(rf.get('/endpoint', data={'page': page}))

    _ = page_number_pagination.paginate_queryset(my_models_queryset, request)

    assert page_number_pagination.page == 1


@pytest.mark.parametrize(
    'per_page,expected_size,expected_has_next',
    (
        (0, 0, True),
        (10, 10, True),
        (5, 5, True),
        (200, 15, False),
    ),
)
@pytest.mark.django_db
def test_cursor_uuid_pagination_paginate_queryset(
    n_models, rf, cursor_uuid_pagination, my_models_queryset, per_page,
    expected_size, expected_has_next,
):
    n_models(15)
    request = Request(rf.get('/endpoint', data={'per_page': per_page}))

    paginated = cursor_uuid_pagination.paginate_queryset(my_models_queryset, request)

    assert len(paginated) == expected_size
    assert cursor_uuid_pagination.has_next == expected_has_next


@pytest.mark.django_db
def test_cursor_uuid_pagination_paginate_before_empty(
    n_models, rf, cursor_uuid_pagination, my_models_queryset,
):
    n_models(15)
    request = Request(rf.get('/endpoint', data={'before': ''}))

    paginated = cursor_uuid_pagination.paginate_queryset(my_models_queryset, request)

    assert len(paginated) == 15
    assert paginated[0].timestamp > paginated[-1].timestamp


@pytest.mark.django_db
def test_cursor_uuid_pagination_paginate_after(
    n_models, rf, cursor_uuid_pagination, my_models_queryset,
):
    n_models(15)
    request = Request(rf.get('/endpoint', data={'after': ''}))

    paginated = cursor_uuid_pagination.paginate_queryset(my_models_queryset, request)

    assert len(paginated) == 15
    assert paginated[0].timestamp < paginated[-1].timestamp


@pytest.mark.parametrize(
    'order,expected_count',
    (
        ('after', 14),
        ('before', 0),
    ),
)
@pytest.mark.django_db
def test_cursor_uuid_pagination_paginate_related_first(
    n_models, rf, cursor_uuid_pagination, my_models_queryset, order, expected_count,
):
    messages = n_models(15)
    request = Request(rf.get('/endpoint', data={order: messages[0].uuid}))

    paginated = cursor_uuid_pagination.paginate_queryset(my_models_queryset, request)

    assert len(paginated) == expected_count


@pytest.mark.parametrize(
    'order,expected_count',
    (
        ('after', 0),
        ('before', 14),
    ),
)
@pytest.mark.django_db
def test_cursor_uuid_pagination_paginate_related_last(
    n_models, rf, cursor_uuid_pagination, my_models_queryset, order, expected_count,
):
    messages = n_models(15)
    request = Request(rf.get('/endpoint', data={order: messages[-1].uuid}))

    paginated = cursor_uuid_pagination.paginate_queryset(my_models_queryset, request)

    assert len(paginated) == expected_count


@pytest.mark.parametrize(
    'index',
    (
        1, 7, 13,
    ),
)
@pytest.mark.django_db
def test_cursor_uuid_pagination_get_paginated_response_after_success_case(
    n_models, rf, cursor_uuid_pagination, my_models_queryset, index,
):
    order = 'after'
    messages = n_models(15)
    msg_uuid = messages[index].uuid
    request = Request(rf.get('/endpoint', data={order: msg_uuid}))

    _ = cursor_uuid_pagination.paginate_queryset(my_models_queryset, request)
    paginated_response = cursor_uuid_pagination.get_paginated_response([])
    boundaries = (
        str(messages[index + 1].uuid),
        str(messages[-1].uuid),
    )
    meta = paginated_response.meta

    assert order in meta
    assert meta[order] == messages[index].uuid
    assert 'after_url' in meta
    assert boundaries[-1] in meta['after_url']
    assert 'before_url' in meta
    assert boundaries[0] in meta['before_url']


@pytest.mark.parametrize(
    'index',
    (
        1, 7, 13,
    ),
)
@pytest.mark.django_db
def test_cursor_uuid_pagination_get_paginated_response_before_success_case(
    n_models, rf, cursor_uuid_pagination, my_models_queryset, index,
):
    order = 'before'
    messages = n_models(15)
    msg_uuid = messages[index].uuid
    request = Request(rf.get('/endpoint', data={order: msg_uuid}))

    _ = cursor_uuid_pagination.paginate_queryset(my_models_queryset, request)
    paginated_response = cursor_uuid_pagination.get_paginated_response([])
    boundaries = (
        str(messages[0].uuid),
        str(messages[index - 1].uuid),
    )
    meta = paginated_response.meta

    assert order in meta
    assert meta[order] == messages[index].uuid
    assert 'after_url' in meta
    assert boundaries[-1] in meta['after_url']
    assert 'before_url' in meta
    assert boundaries[0] in meta['before_url']


@pytest.mark.parametrize(
    'order',
    (
        'after',
        'before',
    ),
)
@pytest.mark.django_db
def test_cursor_uuid_pagination_get_paginated_response_no_uuid_specified_success_case(
    n_models, rf, cursor_uuid_pagination, my_models_queryset, order,
):
    messages = n_models(15)
    request = Request(rf.get('/endpoint', data={order: ''}))

    _ = cursor_uuid_pagination.paginate_queryset(my_models_queryset, request)
    paginated_response = cursor_uuid_pagination.get_paginated_response([])
    boundaries = (
        str(messages[0].uuid),
        str(messages[-1].uuid),
    )
    meta = paginated_response.meta

    assert order in meta
    assert meta[order] is None
    assert 'after_url' in meta
    assert boundaries[1] in meta['after_url']
    assert 'before_url' in meta
    assert boundaries[0] in meta['before_url']



from __future__ import annotations
import typing


if typing.TYPE_CHECKING:
    from rest_framework.permissions import BasePermission
    PermissionClasses = typing.List[BasePermission]
    PermissionClassesMap = typing.Dict[str, PermissionClasses]


def get_permission_classes_from_map(
    action: str,
    permission_classes_map: PermissionClassesMap,
    default_permission_classes: PermissionClasses,
) -> PermissionClasses:
    permission_classes = permission_classes_map.get('default', default_permission_classes)
    return permission_classes_map.get(action, permission_classes)



from __future__ import annotations
from typing import TYPE_CHECKING

from django.conf import settings

if TYPE_CHECKING:
    from typing import Optional
    from restdoctor.utils.custom_types import Prefix, Prefixes, SequenceOrT


def get_api_prefix(default: Optional[Prefix] = '/') -> Optional[Prefix]:
    return get_api_prefixes(default)[0]


def get_api_prefixes(default: Optional[SequenceOrT[Prefix]] = '/') -> Prefixes:
    prefixes = settings.API_PREFIXES or default
    if prefixes is None:
        prefixes = ()
    elif not isinstance(prefixes, tuple):
        prefixes = tuple(prefixes) if isinstance(prefixes, list) else (prefixes,)
    return tuple(
        f'/{prefix.strip("/")}' for prefix in prefixes
    )


def get_api_path_prefix(default: Optional[Prefix] = '/') -> Optional[Prefix]:
    return get_api_path_prefixes(default)[0]


def get_api_path_prefixes(default: Optional[SequenceOrT[Prefix]] = '/') -> Prefixes:
    prefixes = settings.API_PREFIXES or default
    if prefixes is None:
        prefixes = ()
    elif not isinstance(prefixes, tuple):
        prefixes = tuple(prefixes) if isinstance(prefixes, list) else (prefixes,)
    return tuple(
        f'{prefix.strip("/")}/' for prefix in prefixes
    )



from __future__ import annotations

import collections
import functools
import inspect
import logging
import typing

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db.models import Manager, QuerySet
from django.http import Http404
from rest_framework.exceptions import MethodNotAllowed, NotFound

from restdoctor.rest_framework.generics import GenericAPIView
from restdoctor.rest_framework.pagination import PageNumberPagination
from restdoctor.rest_framework.schema import ResourceSchema
from restdoctor.rest_framework.viewsets import GenericViewSet

if typing.TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest
    from rest_framework.pagination import BasePagination
    from rest_framework.response import Response

    from restdoctor.rest_framework.custom_types import (
        ActionMap,
        Handler,
        ResourceActionsMap,
        ResourceExtraAction,
        ResourceHandlersMap,
        ResourceModelsMap,
        ResourceViewsMap,
    )

logger = logging.getLogger(__name__)


def _is_extra_action(attr: typing.Any) -> bool:
    return hasattr(attr, 'mapping')


def get_queryset_model_map(resource_views_map: ResourceViewsMap) -> ResourceModelsMap:
    models_map: ResourceModelsMap = {}

    for resource, viewset_class in resource_views_map.items():
        queryset = getattr(viewset_class, 'queryset', None)
        if isinstance(queryset, (Manager, QuerySet)):
            models_map[resource] = queryset.model
        else:
            models_map[resource] = None

    return models_map


def merge_actions(handlers_actions: typing.Sequence[ActionMap]) -> ActionMap:
    actions = {}
    for handler_actions in handlers_actions:
        actions.update(handler_actions)
    return actions


class ResourceBase:
    schema_class = ResourceSchema

    default_discriminative_value = settings.API_RESOURCE_DEFAULT
    resource_discriminative_param = settings.API_RESOURCE_DISCRIMINATIVE_PARAM
    resource_views_map: ResourceViewsMap = {}
    resource_actions_map: ResourceActionsMap = {}
    resource_handlers_map: ResourceHandlersMap = {}

    resource_discriminate_methods = ['GET']

    @property
    def schema_tags(self) -> typing.List[str]:
        default_view = self.resource_views_map.get(self.default_discriminative_value)
        return getattr(default_view, 'schema_tags', [])

    @classmethod
    def get_resource_handlers(cls, **initkwargs: typing.Any) -> ResourceHandlersMap:
        return {
            resource: viewset.as_view(**initkwargs)
            for resource, viewset in cls.resource_views_map.items()
        }

    @classmethod
    def as_view(cls, **initkwargs: typing.Any) -> typing.Any:
        cls.check_queryset_models()
        resource_handlers_map = cls.get_resource_handlers(**initkwargs)
        view = super().as_view(resource_handlers_map=resource_handlers_map, **initkwargs)  # type: ignore
        return view

    @classmethod
    @functools.lru_cache
    def check_queryset_models(cls) -> bool:
        warnings: typing.List[str] = []
        errors: typing.List[str] = []
        models_map = get_queryset_model_map(cls.resource_views_map)
        current_model = None
        for resource, model in models_map.items():
            if model is None:
                warnings.append(f'Resource {resource} for {cls.__name__} has no queryset.')
            elif current_model is None:
                current_model = model
            elif current_model != model:
                errors.append(
                    (
                        f'Resource {resource} for {cls.__name__} has wrong queryset model '
                        f'{model} instead of {current_model}.'
                    )
                )
        for warning in warnings:
            logger.warning(warning)
        if errors:
            raise ImproperlyConfigured(','.join(errors))
        return True

    def get_discriminant(self, request: WSGIRequest) -> str:
        try:
            if request.api_params.resource_discriminator:
                return request.api_params.resource_discriminator
        except AttributeError:
            pass

        if request.method in self.resource_discriminate_methods:
            discriminant = (
                request.GET.get(self.resource_discriminative_param)
                or self.default_discriminative_value
            )
            if settings.API_RESOURCE_SET_PARAM and (
                discriminant != self.default_discriminative_value
                or settings.API_RESOURCE_SET_PARAM_FOR_DEFAULT
            ):
                request.resource_args = {self.resource_discriminative_param: discriminant}
            return discriminant
        return self.default_discriminative_value

    def dispatch(self, request: WSGIRequest, *args: typing.Any, **kwargs: typing.Any) -> Response:
        discriminant = self.get_discriminant(request)

        resource_dispatch = self.resource_handlers_map.get(discriminant)
        if resource_dispatch:
            return resource_dispatch(request, *args, **kwargs)
        raise Http404()


class ResourceViewSet(ResourceBase, GenericViewSet):
    pagination_class: typing.Optional[BasePagination] = PageNumberPagination

    def __init__(
        self,
        resource_handlers_map: typing.Dict[str, Handler] = None,
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.resource_handlers_map = resource_handlers_map or {}
        for method in self.get_action_methods():
            if not hasattr(self, method):
                setattr(self, method, lambda a: None)

    @classmethod
    def as_view(cls, actions: ActionMap = None, **initkwargs: typing.Any) -> typing.Any:
        initkwargs['actions'] = actions
        return super().as_view(**initkwargs)

    @classmethod
    def get_action_methods(cls) -> typing.List[str]:
        resource_actions_map = collections.defaultdict(set)
        crud_methods_set = {'list', 'retrieve', 'create', 'update', 'partial_update', 'destroy'}
        available_methods_set = set()
        for resource, viewset_class in cls.resource_views_map.items():
            for crud_method in crud_methods_set:
                if hasattr(viewset_class, crud_method):
                    available_methods_set.add(crud_method)
                    resource_actions_map[resource].add(crud_method)
        for resource, name, _ in cls.resources_extra_actions():
            available_methods_set.add(name)
            resource_actions_map[resource].add(name)
        cls.resource_actions_map = resource_actions_map
        return list(available_methods_set)

    @classmethod
    def gen_resources_extra_actions(cls) -> typing.Iterator[ResourceExtraAction]:
        for resource, viewset_class in cls.resource_views_map.items():
            for name, handler in inspect.getmembers(viewset_class, _is_extra_action):
                yield resource, name, handler

    @classmethod
    @functools.lru_cache
    def resources_extra_actions(cls) -> typing.List[ResourceExtraAction]:
        return list(cls.gen_resources_extra_actions())

    @classmethod
    def get_extra_actions(cls) -> typing.List[Handler]:
        return [handler for _, _, handler in cls.resources_extra_actions()]

    def dispatch(self, request: WSGIRequest, *args: typing.Any, **kwargs: typing.Any) -> Response:
        discriminator = self.get_discriminant(request)

        method = request.method.lower()
        action = self.action_map.get(method)
        actions = self.resource_actions_map.get(discriminator)
        if actions is None:
            exc = NotFound()
        elif action in actions:
            return super().dispatch(request, *args, **kwargs)
        else:
            exc = MethodNotAllowed(request.method)
        return self.exception_response(exc, request, *args, **kwargs)

    def exception_response(
        self, exc: Exception, request: WSGIRequest, *args: typing.Any, **kwargs: typing.Any
    ) -> Response:
        request = self.initialize_request(request, *args, **kwargs)
        self.headers = self.default_response_headers
        response = self.handle_exception(exc)
        return self.finalize_response(request, response, *args, **kwargs)


class ResourceView(ResourceBase, GenericAPIView):
    @classmethod
    def as_view(cls, **initkwargs: typing.Any) -> typing.Any:
        view = super().as_view(**initkwargs)
        resource_handlers_map = view.initkwargs.get('resource_handlers_map', {})
        actions = merge_actions(
            [getattr(handler, 'actions', None) for handler in resource_handlers_map.values()]
        )
        if actions:
            view.actions = actions
        return view



from __future__ import annotations

from rest_framework.viewsets import ViewSetMixin as BaseViewSetMixin

from restdoctor.rest_framework.mixins import (
    RetrieveModelMixin,
    ListModelMixin,
    CreateModelMixin,
    UpdateModelMixin,
    DestroyModelMixin,
)
from restdoctor.rest_framework.views import SerializerClassMapApiView


class ViewSetMixin(BaseViewSetMixin):
    pass


class GenericViewSet(ViewSetMixin, SerializerClassMapApiView):
    pass


class ListModelViewSet(
    ListModelMixin,
    GenericViewSet,
):
    pass


class ReadOnlyModelViewSet(
    RetrieveModelMixin,
    ListModelViewSet,
):
    pass


class CreateUpdateReadModelViewSet(
    CreateModelMixin,
    RetrieveModelMixin,
    UpdateModelMixin,
    ListModelViewSet,
):
    pass


class ModelViewSet(
    CreateModelMixin,
    RetrieveModelMixin,
    UpdateModelMixin,
    DestroyModelMixin,
    ListModelViewSet,
):
    pass



from __future__ import annotations

import pytest
from django.core.exceptions import ImproperlyConfigured
from django.http import Http404
from django.test import RequestFactory
from rest_framework.status import HTTP_200_OK, HTTP_404_NOT_FOUND, HTTP_405_METHOD_NOT_ALLOWED

from restdoctor.rest_framework.resources import (
    ResourceView,
    ResourceViewSet,
    get_queryset_model_map,
)
from tests.test_unit.stubs import (
    ComplexResourceViewSet,
    ModelA,
    ModelAView,
    ModelAViewSet,
    ModelAWithMixinViewSet,
    ModelBViewSet,
    NoneViewSet,
)


@pytest.mark.django_db
@pytest.mark.parametrize(
    ('resource_views_map', 'expected_model_map'),
    [
        ({'one': ModelAViewSet}, {'one': ModelA}),
        ({'one': ModelAWithMixinViewSet}, {'one': ModelA}),
        ({'one': NoneViewSet}, {'one': None}),
    ],
)
def test_get_queryset_model_map_success_case(resource_views_map, expected_model_map):
    model_map = get_queryset_model_map(resource_views_map)

    assert model_map == expected_model_map


@pytest.mark.django_db
@pytest.mark.parametrize(
    ('base_class', 'resource_views_map'),
    [
        (ResourceViewSet, {'one': ModelAViewSet}),
        (ResourceViewSet, {'one': ModelAViewSet, 'another': ModelAWithMixinViewSet}),
        (ResourceViewSet, {'one': ModelAViewSet, 'another': NoneViewSet}),
        (ResourceViewSet, {'one': NoneViewSet}),
        (ResourceView, {'one': ModelAViewSet}),
        (ResourceView, {'one': ModelAViewSet, 'another': ModelAWithMixinViewSet}),
        (ResourceView, {'one': ModelAViewSet, 'another': NoneViewSet}),
        (ResourceView, {'one': NoneViewSet}),
    ],
)
def test_check_queryset_models_success_case(base_class, resource_views_map):
    resource_class = type(
        'TestResourceViewSet', (base_class,), {'resource_views_map': resource_views_map}
    )

    result = resource_class.check_queryset_models()

    assert result is True


@pytest.mark.django_db
@pytest.mark.parametrize(
    ('base_class', 'resource_views_map'),
    [
        (ResourceViewSet, {'one': ModelAViewSet, 'another': ModelBViewSet}),
        (ResourceView, {'one': ModelAViewSet, 'another': ModelBViewSet}),
    ],
)
def test_check_queryset_models_fail_case(base_class, resource_views_map):
    resource_class = type('TestResource', (base_class,), {'resource_views_map': resource_views_map})

    with pytest.raises(ImproperlyConfigured):
        resource_class.check_queryset_models()


@pytest.mark.django_db
def test_resource_viewset_dispatch_no_default_discriminator_fail_case(resource_viewset_dispatch):
    resource_discriminator = 'one'
    view_func, _ = resource_viewset_dispatch(
        resource_discriminator, ModelAViewSet, actions={'get': 'retrieve'}
    )
    request = RequestFactory().get('/')

    response = view_func(request)

    assert response.status_code == HTTP_404_NOT_FOUND


@pytest.mark.django_db
def test_resource_viewset_dispatch_wrong_discriminator_fail_case(resource_viewset_dispatch):
    resource_discriminator = 'one'
    view_func, _ = resource_viewset_dispatch(
        resource_discriminator, ModelAViewSet, actions={'get': 'retrieve'}
    )
    request = RequestFactory().get('/', {'view_type': f'NOT_{resource_discriminator}'})

    response = view_func(request)

    assert response.status_code == HTTP_404_NOT_FOUND


@pytest.mark.django_db
def test_resource_viewset_dispatch_success_case(resource_viewset_dispatch):
    resource_discriminator = 'one'
    view_func, mocked_dispatch = resource_viewset_dispatch(
        resource_discriminator, ModelAViewSet, actions={'get': 'retrieve'}
    )
    request = RequestFactory().get('/', {'view_type': resource_discriminator})

    view_func(request)

    assert mocked_dispatch.called_once()


@pytest.mark.django_db
def test_resource_view_dispatch_no_default_discriminator_fail_case(resource_view_dispatch):
    resource_discriminator = 'one'
    view_func, _ = resource_view_dispatch(resource_discriminator, ModelAView)
    request = RequestFactory().get('/')

    with pytest.raises(Http404):
        view_func(request)


@pytest.mark.django_db
def test_resource_view_dispatch_wrong_discriminator_fail_case(resource_view_dispatch):
    resource_discriminator = 'one'
    view_func, _ = resource_view_dispatch(resource_discriminator, ModelAView)
    request = RequestFactory().get('/', {'view_type': f'NOT_{resource_discriminator}'})

    with pytest.raises(Http404):
        view_func(request)


@pytest.mark.django_db
def test_resource_view_dispatch_success_case(resource_view_dispatch):
    resource_discriminator = 'one'
    view_func, mocked_dispatch = resource_view_dispatch(resource_discriminator, ModelAView)
    request = RequestFactory().get('/', {'view_type': resource_discriminator})

    view_func(request)

    assert mocked_dispatch.called_once()


@pytest.mark.parametrize(
    ('accept', 'view_type', 'expected_discriminant'),
    [
        ('application/vnd.vendor.v1', '', 'common'),
        ('application/vnd.vendor.v1', 'extended', 'extended'),
        ('application/vnd.vendor.v1-extended', '', 'extended'),
        ('application/vnd.vendor.v1-extended', 'common', 'extended'),
        ('application/vnd.vendor.v2', '', 'common'),
    ],
)
@pytest.mark.django_db
def test_get_discriminant_for_get_method(
    accept, view_type, expected_discriminant, settings, client, api_prefix, get_discriminant_spy
):
    settings.API_VERSIONS = {'v1': 'tests.stubs.api.v1_urls'}
    settings.API_RESOURCE_DEFAULT = 'common'

    client.get(f'/{api_prefix}mymodel/', HTTP_ACCEPT=accept, data={'view_type': view_type})

    assert get_discriminant_spy.spy_return == expected_discriminant


@pytest.mark.parametrize(
    ('accept', 'expected_discriminant'),
    [
        ('application/vnd.vendor.v1', 'common'),
        ('application/vnd.vendor.v1-common', 'common'),
        ('application/vnd.vendor.v1-extended', 'extended'),
    ],
)
@pytest.mark.django_db
def test_get_discriminant_for_post_method(
    accept, expected_discriminant, settings, client, api_prefix, get_discriminant_spy
):
    settings.API_VERSIONS = {'v1': 'tests.stubs.api.v1_urls'}
    settings.API_RESOURCE_DEFAULT = 'common'

    client.post(f'/{api_prefix}mymodel/', HTTP_ACCEPT=accept)

    assert get_discriminant_spy.spy_return == expected_discriminant


@pytest.mark.parametrize(
    ('discriminator', 'expected_status_code'),
    [
        ('read_write', HTTP_200_OK),
        ('read_only', HTTP_405_METHOD_NOT_ALLOWED),
        ('default', HTTP_404_NOT_FOUND),
    ],
)
def test_resource_view_dispatch_complex_case_raises(mocker, discriminator, expected_status_code):
    mocked_get_discriminant = mocker.patch.object(ComplexResourceViewSet, 'get_discriminant')
    mocked_get_discriminant.return_value = discriminator
    view_func = ComplexResourceViewSet.as_view(actions={'post': 'update'})
    request = RequestFactory().post('/')

    response = view_func(request)

    assert response.status_code == expected_status_code



from __future__ import annotations
import functools
import typing

from django.core.exceptions import ImproperlyConfigured

from restdoctor.rest_framework.schema.constants import ACTION_CODES_MAP

if typing.TYPE_CHECKING:
    from rest_framework.generics import GenericAPIView
    from restdoctor.rest_framework.custom_types import Action
    from restdoctor.rest_framework.schema.custom_types import (
        ActionDescription, CodeActionSchemaTuple, ResponseCode, OpenAPISchema,
    )

    ActionSchemaTuple = typing.Tuple[ActionDescription, typing.Optional[OpenAPISchema]]
    ActionSchema = typing.Union[ActionDescription, ActionSchemaTuple]
    OptionalActionSchema = typing.Optional[ActionSchema]
    CodeSchemasMap = typing.Dict[ResponseCode, OptionalActionSchema]
    ActionCodesMap = typing.Dict[Action, CodeSchemasMap]


@functools.lru_cache()
def get_action(path: str, method: str, view: GenericAPIView) -> str:
    action = getattr(view, 'action', None)
    if action:
        return action
    action_map = getattr(view, 'action_map', None) or {}
    method_name = method.lower()
    return action_map.get(method_name, method_name)


def normalize_action_schema(code: str, action_schema: ActionSchema) -> CodeActionSchemaTuple:
    if isinstance(action_schema, str):
        return code, action_schema, None
    description, schema = action_schema
    return code, description, schema


def get_action_map_kwargs(
    action: str, action_codes_map: typing.Optional[ActionCodesMap],
) -> typing.List[typing.Tuple[str, ActionCodesMap]]:
    kwargs_variants = []
    kwargs_use_defaults = True
    if isinstance(action_codes_map, dict):
        kwargs_variants.extend([
            (action, action_codes_map),
            ('default', action_codes_map),
        ])
        if action_codes_map.get('default', {}) is None:
            kwargs_use_defaults = False
    if kwargs_use_defaults:
        kwargs_variants.extend([
            (action, ACTION_CODES_MAP),
            ('default', ACTION_CODES_MAP),
        ])
    return kwargs_variants


def get_action_code_schemas_from_map(
    action: str, action_codes_map: ActionCodesMap,
) -> typing.Iterator[typing.Tuple[str, OptionalActionSchema]]:
    action_codes = action_codes_map.get(action, {})

    if isinstance(action_codes, set):
        raise ImproperlyConfigured(
            f'schema_action_codes_map item should be dict for action {action}: {action_codes}')

    for code, action_schema in action_codes.items():
        yield code, action_schema



import pytest
from rest_framework.fields import CharField

from restdoctor.rest_framework.serializers import extend_meta_fields

from tests.test_unit.test_serializers.test_meta_inheritance.stubs import (
    ParentModelSerializer, ParentSerializer, SerializerMixinA, SerializerMixinB,
)


@pytest.mark.parametrize(
    'parent_serializer_cls',
    [
        ParentSerializer,
        ParentModelSerializer,
    ],
)
def test_with_all_parents_explicit(parent_serializer_cls):
    class ChildSerializer(SerializerMixinA, SerializerMixinB, parent_serializer_cls):
        class Meta:
            fields = extend_meta_fields(
                SerializerMixinA, 'field_child_3', parent_serializer_cls, SerializerMixinB, 'field_child_1')

        field_child_1 = CharField()
        field_child_2 = CharField()
        field_child_3 = CharField()

    assert ChildSerializer.Meta.fields == [
        'field_mixin_a', 'field_child_3', 'field_parent_1', 'field_parent_2', 'field_mixin_b', 'field_child_1',
    ]


def test_with_all_parents_implicit():
    class ChildSerializer(SerializerMixinA, SerializerMixinB, ParentModelSerializer):
        class Meta:
            fields = extend_meta_fields('field_child_3', 'field_child_1')

        field_child_1 = CharField()
        field_child_2 = CharField()
        field_child_3 = CharField()

    assert ChildSerializer.Meta.fields == [
        'field_mixin_a', 'field_mixin_b', 'field_parent_1', 'field_parent_2', 'field_child_3', 'field_child_1',
    ]


def test_with_some_parents():
    class ChildSerializer(SerializerMixinA, SerializerMixinB, ParentModelSerializer):
        class Meta:
            fields = extend_meta_fields(SerializerMixinA, ParentModelSerializer, 'field_child')

        field_child = CharField()

    assert ChildSerializer.Meta.fields == ['field_mixin_a', 'field_parent_1', 'field_parent_2', 'field_child']


def test_without_extend():
    class ChildSerializer(SerializerMixinA, SerializerMixinB, ParentModelSerializer):
        class Meta:
            fields = ['field_child_3', 'field_child_1']

        field_child_1 = CharField()
        field_child_2 = CharField()
        field_child_3 = CharField()

    assert ChildSerializer.Meta.fields == ['field_child_3', 'field_child_1']



import pytest

from restdoctor.rest_framework.schema import RefsSchemaGenerator
from tests.test_unit.test_schema.stubs import DefaultViewSet


@pytest.mark.parametrize(
    'generator,expected_codes',
    (
        (None, ['200']),
        (RefsSchemaGenerator(), ['200', '400', '404']),
    ),
)
def test_errors_handlers_sc(
    get_create_view_func, generator, expected_codes,
):
    create_view = get_create_view_func('test', DefaultViewSet, 'test')

    view = create_view('/test/123/', 'GET')
    view.schema.generator = generator
    operation = view.schema.get_operation('/test/123/', 'GET')

    print(generator, operation)
    assert list(operation['responses'].keys()) == expected_codes



"""
Django settings for stub_site project.

Generated by 'django-admin startproject' using Django 3.1.1.

For more information on this file, see
https://docs.djangoproject.com/en/3.1/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/3.1/ref/settings/
"""

from pathlib import Path

# Build paths inside the project like this: BASE_DIR / 'subdir'.
from pytz import timezone

BASE_DIR = Path(__file__).resolve().parent.parent


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/3.1/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = 'p+i3ew^n1wi6c=x0lc!s1&9!#4sv_$g&d5p77!#fk&$q3_bqcc'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []


# Application definition

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'rest_framework',
    'restdoctor',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'restdoctor.django.middleware.api_selector.ApiSelectorMiddleware',
]

ROOT_URLCONF = 'tests.stubs.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

# Database
# https://docs.djangoproject.com/en/3.1/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        #'NAME': ':memory:',
    }
}

MIGRATION_MODULES = {'restdoctor': 'tests.stubs.migrations'}

# Password validation
# https://docs.djangoproject.com/en/3.1/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]


# Internationalization
# https://docs.djangoproject.com/en/3.1/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'Europe/Moscow'
TIME_ZONE_TZ = timezone('Europe/Moscow')

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/3.1/howto/static-files/

STATIC_URL = '/static/'

REST_FRAMEWORK = {
    'EXCEPTION_HANDLER': 'restdoctor.rest_framework.exception_handlers.exception_handler',
}

API_FALLBACK_VERSION = 'fallback'
API_FALLBACK_FOR_APPLICATION_JSON_ONLY = False
API_DEFAULT_VERSION = 'v1'
API_DEFAULT_FORMAT = 'full'
API_PREFIXES = ('/api',)
API_FORMATS = ('full', 'compact')
API_RESOURCE_DISCRIMINATIVE_PARAM = 'view_type'
API_RESOURCE_DEFAULT = 'common'
API_RESOURCE_SET_PARAM = False
API_RESOURCE_SET_PARAM_FOR_DEFAULT = False
API_FALLBACK_URLCONF = 'tests.stubs.api.fallback_urls'
API_V1_URLCONF = 'tests.stubs.api.v1_urls'
API_VERSIONS = {
    'fallback': 'tests.stubs.api.fallback_urls',
    'v1': 'tests.stubs.api.v1_urls',
}



from __future__ import annotations
import typing as t

from django.db import models
from rest_framework.parsers import BaseParser
from rest_framework.renderers import BaseRenderer
from rest_framework.response import Response
from rest_framework.routers import DynamicRoute, Route
from rest_framework.viewsets import ViewSet

Action = str

ActionMap = t.Dict[Action, str]
Handler = t.Callable[..., Response]
ResourceExtraAction = t.Tuple[str, str, Handler]

RouteOrDynamicRoute = t.Union[Route, DynamicRoute]
RouteOrDynamicRouteList = t.List[RouteOrDynamicRoute]

Parsers = t.Sequence[BaseParser]
OptionalParser = t.Optional[BaseParser]

Renderers = t.Sequence[BaseRenderer]
OptionalRenderer = t.Optional[BaseRenderer]

ResourceMapElement = t.TypeVar('ResourceMapElement')
ResourceMap = t.Dict[str, ResourceMapElement]

ResourceViewsMap = ResourceMap[t.Type[ViewSet]]
ResourceActionsMap = ResourceMap[t.Set[str]]
ResourceHandlersMap = ResourceMap[Handler]
ResourceModelsMap = ResourceMap[t.Optional[models.Model]]



from __future__ import annotations

import typing as t

from rest_framework.fields import Field
from rest_framework.schemas.openapi import SchemaGenerator
from rest_framework.serializers import BaseSerializer

OpenAPISchema = t.Dict[str, 'OpenAPISchema']  # type: ignore
LocalRefs = t.Dict[t.Tuple[str, ...], t.Any]

ResponseCode = str
ActionDescription = str

CodeActionSchemaTuple = t.Tuple[ResponseCode, ActionDescription, t.Optional[OpenAPISchema]]  # type: ignore
CodeDescriptionTuple = t.Tuple[ResponseCode, ActionDescription]


class ViewSchemaProtocol(t.Protocol):
    generator: t.Optional[SchemaGenerator] = None
    serializer_schema: SerializerSchemaProtocol
    field_schema: FieldSchemaProtocol

    def get_field_schema(self, field: Field) -> OpenAPISchema:
        return self.field_schema.get_field_schema(field)

    def get_serializer_schema(
        self,
        serializer: BaseSerializer,
        write_only: bool = True,
        read_only: bool = True,
        required: bool = True,
    ) -> OpenAPISchema:
        return self.serializer_schema.get_serializer_schema(
            serializer, write_only=write_only, read_only=read_only, required=required
        )

    def map_field(self, field: Field) -> t.Optional[OpenAPISchema]:
        return self.field_schema.map_field(field)

    def get_field_description(self, field: Field) -> t.Optional[str]:
        return self.field_schema.get_field_description(field)

    def map_field_validators(self, field: Field, schema: OpenAPISchema) -> None:
        return self.field_schema.map_field_validators(field, schema)

    def map_serializer(
        self,
        serializer: BaseSerializer,
        write_only: bool = True,
        read_only: bool = True,
        required: bool = True,
    ) -> OpenAPISchema:
        return self.serializer_schema.map_serializer(
            serializer, write_only=write_only, read_only=read_only, required=required
        )

    def map_query_serializer(self, serializer: BaseSerializer) -> t.List[OpenAPISchema]:
        return self.serializer_schema.map_query_serializer(serializer)


class SerializerSchemaProtocol(t.Protocol):
    view_schema: ViewSchemaProtocol

    def get_serializer_schema(
        self,
        serializer: BaseSerializer,
        write_only: bool = True,
        read_only: bool = True,
        required: bool = True,
    ) -> OpenAPISchema:
        ...

    def map_serializer(
        self,
        serializer: BaseSerializer,
        write_only: bool = True,
        read_only: bool = True,
        required: bool = True,
    ) -> OpenAPISchema:
        ...

    def map_query_serializer(self, serializer: BaseSerializer) -> t.List[OpenAPISchema]:
        ...


class FieldSchemaProtocol(t.Protocol):
    view_schema: ViewSchemaProtocol

    def get_field_schema(self, field: Field) -> OpenAPISchema:
        ...

    def map_field(self, field: Field) -> t.Optional[OpenAPISchema]:
        ...

    def get_field_description(self, field: Field) -> t.Optional[str]:
        ...

    def map_field_validators(self, field: Field, schema: OpenAPISchema) -> None:
        ...



from __future__ import annotations
import typing

if typing.TYPE_CHECKING:
    from rest_framework.response import Response


def unpacked_error_message(response: Response) -> str:
    return response.json()['errors'][0]['message']


def get_full_class_name(obj: typing.Any) -> str:
    module = obj.__class__.__module__
    if module is None or module == str.__class__.__module__:
        return obj.__class__.__name__
    return module + '.' + obj.__class__.__name__



import pytest

from tests.test_unit.test_schema.stubs import ListViewSetWithRequestSerializer, ListViewSetWithoutRequestSerializer


@pytest.mark.parametrize(
    'viewset,expected_parameters',
    (
        (
            ListViewSetWithRequestSerializer, [
                {
                    'in': 'query',
                    'name': 'filter_uuid_field',
                    'required': False,
                    'schema': {
                        'type': 'string',
                        'format': 'uuid',
                    },
                },
                {
                    'in': 'query',
                    'name': 'filter_field',
                    'required': True,
                    'schema': {
                        'type': 'string',
                        'nullable': True,
                        'description': 'Help text',
                    },
                },
            ],
        ),
        (ListViewSetWithoutRequestSerializer, []),
    ),
)
def test_schema_for_list_viewset(get_create_view_func, viewset, expected_parameters):
    create_view = get_create_view_func('test', viewset, 'test')

    view = create_view('/test/', 'GET')
    operation = view.schema.get_operation('/test/', 'GET')

    assert operation['parameters'] == expected_parameters


@pytest.mark.parametrize(
    'viewset,expected_parameters',
    (
        (
            ListViewSetWithRequestSerializer, [
                {
                    'in': 'query',
                    'name': 'filter_uuid_field',
                    'required': False,
                    'schema': {
                        'type': 'string',
                        'format': 'uuid',
                    },
                },
                {
                    'in': 'query',
                    'name': 'filter_field',
                    'required': True,
                    'schema': {
                        'type': 'string',
                        'nullable': True,
                        'description': 'Help text',
                    },
                },
            ],
        ),
        (ListViewSetWithoutRequestSerializer, []),
    ),
)
def test_get_request_serializer_filter_parameters(get_create_view_func, viewset, expected_parameters):
    list_view = get_create_view_func('test', viewset, 'test')

    view = list_view('/test/', 'GET')
    parameters = view.schema.get_request_serializer_filter_parameters('/test/', 'GET')

    assert parameters == expected_parameters



from __future__ import annotations
import typing

from django.conf import settings

from restdoctor.rest_framework.serializers import EmptySerializer

if typing.TYPE_CHECKING:
    from rest_framework.serializers import BaseSerializer
    from restdoctor.utils.custom_types import GenericContext

    SerializerType = typing.Type[BaseSerializer]
    SerializerClassMapItem = typing.Dict[str, SerializerType]
    SerializerClassMap = typing.Dict[str, typing.Union[SerializerType, SerializerClassMapItem]]


def get_from_serializer(
    serializer_class: BaseSerializer,
    data: GenericContext,
    instance: typing.Any = None,
    raise_exception: bool = True,
    **kwargs: typing.Any,
) -> GenericContext:
    serializer = get_serializer(serializer_class, instance, data, **kwargs)
    if serializer.is_valid(raise_exception=raise_exception):
        return serializer.validated_data
    return {}


def get_serializer(
    serializer_class: SerializerType,
    instance: typing.Any = None,
    data: typing.Optional[GenericContext] = None,
    **kwargs: typing.Any,
) -> BaseSerializer:
    return serializer_class(instance=instance, data=data, **kwargs)


def modify_instance(
    instance: typing.Any,
    update_data: GenericContext,
) -> bool:
    modified = False
    for key, value in update_data.items():
        old_value = getattr(instance, key)
        if old_value != value:
            setattr(instance, key, value)
            modified = True
    return modified


def update_instance(
    instance: typing.Any,
    update_data: GenericContext,
    save_if_modified: bool = True,
) -> bool:
    modified = modify_instance(instance, update_data)
    if modified and save_if_modified:
        instance.save()

        if getattr(instance, '_prefetched_objects_cache', None):
            instance._prefetched_objects_cache = {}
    return modified


def get_serializer_class_from_map(
    action: str,
    stage: str,
    serializer_class_map: SerializerClassMap,
    default_class: SerializerType = EmptySerializer,
    use_default: bool = True,
    api_format: str = None,
) -> BaseSerializer:
    api_format = api_format or settings.API_DEFAULT_FORMAT
    if use_default:
        serializer_class = serializer_class_map.get('default', default_class)
        serializer_class = serializer_class_map.get(f'default.{api_format}', serializer_class)
    else:
        serializer_class = EmptySerializer

    action_class_map = serializer_class_map.get(action)
    if isinstance(action_class_map, dict):
        serializer_class = action_class_map.get(stage, serializer_class)
        serializer_class = action_class_map.get(f'{stage}.{api_format}', serializer_class)
    return serializer_class



from __future__ import annotations

import typing

from django.conf import settings
from rest_framework.fields import Field
from rest_framework.pagination import BasePagination
from rest_framework.schemas.openapi import AutoSchema, SchemaGenerator
from rest_framework.schemas.utils import is_list_view
from rest_framework.serializers import BaseSerializer

from restdoctor.rest_framework.pagination.mixins import SerializerClassPaginationMixin
from restdoctor.rest_framework.schema.custom_types import ViewSchemaProtocol
from restdoctor.rest_framework.schema.fields import FieldSchema
from restdoctor.rest_framework.schema.serializers import SerializerSchema
from restdoctor.rest_framework.schema.utils import (
    get_action,
    get_action_code_schemas_from_map,
    get_action_map_kwargs,
    normalize_action_schema,
)
from restdoctor.rest_framework.serializers import EmptySerializer
from restdoctor.rest_framework.views import SerializerClassMapApiView

if typing.TYPE_CHECKING:
    from restdoctor.rest_framework.schema.custom_types import (
        CodeActionSchemaTuple,
        CodeDescriptionTuple,
        OpenAPISchema,
    )


class RestDoctorSchema(ViewSchemaProtocol, AutoSchema):
    def __init__(
        self, generator: SchemaGenerator = None, *args: typing.Any, **kwargs: typing.Any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.generator = generator
        self.field_schema = FieldSchema(self)
        self.serializer_schema = SerializerSchema(self)

    def map_renderers(self, *args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        vendor = getattr(settings, 'API_VENDOR_STRING', 'vendor').lower()
        media_types = [f'application/vnd.{vendor}']
        return media_types

    def get_operation(self, path: str, method: str) -> OpenAPISchema:
        operation = super().get_operation(path, method)
        operation['parameters'] += self.get_request_serializer_filter_parameters(path, method)
        operation['tags'] = self.get_tags(path, method)

        return operation

    def get_request_serializer_filter_parameters(
        self, path: str, method: str
    ) -> typing.List[OpenAPISchema]:
        if not is_list_view(path, method, self.view):
            return []

        parameters = []
        request_serializer_class = self.view.get_request_serializer_class(use_default=False)
        request_serializer = request_serializer_class()
        if not isinstance(request_serializer, EmptySerializer):
            for field in request_serializer.fields.values():
                field_schema = self.field_schema.get_field_schema(field)
                parameters.append(
                    {
                        'name': field.field_name,
                        'required': field.required,
                        'in': 'query',
                        'schema': field_schema,
                    }
                )
        return parameters

    def get_action_name(self, path: str, method: str) -> str:
        action = get_action(path, method, self.view)
        if is_list_view(path, method, self.view):
            return 'list'
        elif action not in self.method_mapping:
            return action.lower()
        else:
            return self.method_mapping[method.lower()].lower()

    def get_object_name_by_view_class_name(
        self, clean_suffixes: typing.Sequence[str] = None
    ) -> str:
        clean_suffixes = clean_suffixes or []
        object_name = self.view.__class__.__name__
        for suffix in sorted(clean_suffixes, key=len, reverse=True):
            if object_name.endswith(suffix):
                object_name = object_name[: -len(suffix)]
                break
        return object_name

    def get_object_name(self, path: str, method: str, action_name: str) -> str:
        action = get_action(path, method, self.view)

        # Try to deduce the ID from the view's model
        model = getattr(getattr(self.view, 'queryset', None), 'model', None)
        if model is not None:
            return model.__name__

        # Try with the serializer class name
        if isinstance(self.view, SerializerClassMapApiView):
            serializer_class = self.view.get_serializer_class('response', action)
            if serializer_class:
                object_name = serializer_class.__name__
                if object_name.endswith('Serializer'):
                    object_name = object_name[:-10]
                return object_name

        object_name = self.get_object_name_by_view_class_name(
            clean_suffixes=['APIView', 'View', 'ViewSet']
        )

        # Due to camel-casing of classes and `action` being lowercase, apply title in order to find if action truly
        # comes at the end of the name
        if object_name.endswith(action_name.title()):  # ListView, UpdateAPIView, ThingDelete ...
            object_name = object_name[: -len(action_name)]

        return object_name

    def get_operation_id(self, path: str, method: str) -> str:
        action_name = self.get_action_name(path, method)
        object_name = self.get_object_name(path, method, action_name)

        if action_name == 'list' and not object_name.endswith(
            's'
        ):  # listThings instead of listThing
            object_name += 's'

        return action_name + object_name

    def get_pagination_parameters(self, path: str, method: str) -> typing.List[OpenAPISchema]:
        if not is_list_view(path, method, self.view):
            return []

        paginator: BasePagination = self.get_paginator()
        if not paginator:
            return []

        if isinstance(paginator, SerializerClassPaginationMixin):
            serializer_class = paginator.get_request_serializer_class()
        else:
            serializer_class = getattr(paginator, 'serializer_class', None)
        if serializer_class:
            return self.serializer_schema.map_query_serializer(serializer_class())
        return paginator.get_schema_operation_parameters(self.view)

    def get_serializer(
        self, path: str, method: str, stage: str, api_format: str = None
    ) -> typing.Optional[BaseSerializer]:
        view = self.view
        action = get_action(path, method, view)

        serializer_class = None
        if issubclass(view.__class__, SerializerClassMapApiView):
            serializer_class = view.get_serializer_class(stage, action, api_format)

        if serializer_class is None:
            serializer = view.get_serializer()
        else:
            serializer = serializer_class()

        return serializer

    def get_request_body_schema(
        self, path: str, method: str, api_format: str = None
    ) -> OpenAPISchema:
        serializer = self.get_serializer(path, method, 'request', api_format=api_format)

        if serializer is None:
            return {}

        return self.map_serializer(serializer, read_only=False, required=(method != 'PATCH'))

    def get_request_body(self, path: str, method: str) -> OpenAPISchema:
        if method not in ('PUT', 'PATCH', 'POST'):
            return {}

        self.request_media_types = self.map_parsers(path, method)
        schema = self.get_request_body_schema(path, method)

        return {'content': {ct: {'schema': schema} for ct in self.request_media_types}}

    def get_item_schema(
        self, path: str, method: str, api_format: str = None
    ) -> typing.Optional[OpenAPISchema]:
        serializer = self.get_serializer(path, method, 'response', api_format=api_format)

        if serializer is None:
            return None

        return self.serializer_schema.map_serializer(serializer, write_only=False)

    def get_content_schema(
        self, response_schema: OpenAPISchema, description: str = ''
    ) -> OpenAPISchema:
        content_schema: OpenAPISchema = {'description': description}
        if response_schema:
            content_schema['content'] = {
                content_type: {'schema': response_schema}
                for content_type in self.response_media_types
            }
        return content_schema

    def get_response_schema(self, path: str, method: str, api_format: str = None) -> OpenAPISchema:
        response_schema: OpenAPISchema = {'type': 'object', 'properties': {}}

        item_schema = self.get_item_schema(path, method, api_format)

        if is_list_view(path, method, self.view):
            response_schema['properties']['data'] = {'type': 'array', 'items': item_schema}
            paginator = self.get_paginator()
            if paginator:
                response_schema = paginator.get_paginated_response_schema(response_schema)
        else:
            response_schema['properties']['data'] = item_schema

        return response_schema

    def get_paginator(self) -> typing.Optional[BasePagination]:
        pagination_class = getattr(self.view, 'pagination_class', None)
        if pagination_class:
            return pagination_class(view_schema=self)
        return None

    def get_action_code_description(self, path: str, method: str) -> CodeDescriptionTuple:
        for code, description, _ in self.get_action_code_schemas(path, method):
            if code < '400':
                return code, description
        return '200', 'Успешный запрос.'

    def get_action_code_schemas(
        self, path: str, method: str
    ) -> typing.Iterator[CodeActionSchemaTuple]:
        success_codes_seen: typing.Set[str] = set()
        error_codes_seen: typing.Set[str] = set()
        action = get_action(path, method, self.view)
        kwargs_variants = get_action_map_kwargs(
            action, getattr(self.view, 'schema_action_codes_map', None)
        )

        for kwargs_variant in kwargs_variants:
            for code, action_schema in get_action_code_schemas_from_map(*kwargs_variant):
                if code > '400':
                    codes_seen_set = error_codes_seen
                else:
                    codes_seen_set = success_codes_seen
                if code not in codes_seen_set:
                    codes_seen_set.add(code)
                    if action_schema is not None:
                        yield normalize_action_schema(code, action_schema)

    def get_content_schema_by_type(self, path: str, method: str, schema_type: str) -> OpenAPISchema:
        content_schema = {}

        if schema_type == 'responses':
            schema_method_name = 'get_response_schema'
        else:
            schema_method_name = 'get_request_body_schema'

        vendor = getattr(settings, 'API_VENDOR_STRING', 'vendor').lower()
        default_content_type = f'application/vnd.{vendor}'

        schema_method = getattr(self, schema_method_name)
        default_schema = schema_method(path, method)

        content_schema[default_content_type] = {'schema': default_schema}

        if not self.generator:
            return content_schema

        version_content_type = f'{default_content_type}.{self.generator.api_version}'

        schema_method = getattr(self.view.schema, schema_method_name)

        for api_format in self.generator.api_formats:
            if api_format != self.generator.api_default_format:
                resource_schema = schema_method(path, method, api_format=api_format)
                if resource_schema != default_schema:
                    content_type = f'{version_content_type}.{api_format}'
                    content_schema[content_type] = {'schema': resource_schema}

        return content_schema

    def get_responses(self, path: str, method: str) -> OpenAPISchema:
        schema: OpenAPISchema = {}
        self.response_media_types = self.map_renderers(path, method)

        for code, description, action_schema in self.get_action_code_schemas(path, method):
            if action_schema is None:
                schema[code] = {
                    'description': description,
                    'content': self.get_content_schema_by_type(path, method, 'responses'),
                }
            else:
                if '$ref' not in action_schema or self.generator:
                    schema[code] = self.get_content_schema(action_schema, description=description)

        return schema

    def get_tags(self, path: str, method: str) -> typing.List[str]:
        view = self.view

        schema_tags = getattr(view, 'schema_tags', [])
        if schema_tags:
            return view.schema_tags

        schema_tags = getattr(self, '_tags', [])
        if schema_tags:
            return view.schema_tags

        # First element of a specific path could be valid tag. This is a fallback solution.
        # PUT, PATCH, GET(Retrieve), DELETE:        /user_profile/{id}/       tags = [user-profile]
        # POST, GET(List):                          /user_profile/            tags = [user-profile]
        if path.startswith('/'):
            path = path[1:]

        return [path.split('/')[0].replace('_', '-')]

    def _get_action_name(self, path: str, method: str) -> str:
        return self.get_action_name(path, method)

    def _get_object_name_by_view_class_name(
        self, clean_suffixes: typing.Sequence[str] = None
    ) -> str:
        return self.get_object_name_by_view_class_name(clean_suffixes=clean_suffixes)

    def _get_object_name(self, path: str, method: str, action_name: str) -> str:
        return self.get_object_name(path, method, action_name)

    def _get_operation_id(self, path: str, method: str) -> str:
        return self.get_operation_id(path, method)

    def _get_pagination_parameters(self, path: str, method: str) -> typing.List[OpenAPISchema]:
        return self.get_pagination_parameters(path, method)

    def _get_serializer(
        self, path: str, method: str, stage: str, api_format: str = None
    ) -> typing.Optional[BaseSerializer]:
        return self.get_serializer(path, method, stage, api_format=api_format)

    def _get_request_body_schema(
        self, path: str, method: str, api_format: str = None
    ) -> OpenAPISchema:
        return self.get_request_body_schema(path, method, api_format=api_format)

    def _get_request_body(self, path: str, method: str) -> OpenAPISchema:
        return self.get_request_body(path, method)

    def _get_item_schema(
        self, path: str, method: str, api_format: str = None
    ) -> typing.Optional[OpenAPISchema]:
        return self.get_item_schema(path, method, api_format=api_format)

    def _get_content_schema(
        self, response_schema: OpenAPISchema, description: str = ''
    ) -> OpenAPISchema:
        return self.get_content_schema(response_schema, description=description)

    def _get_response_schema(self, path: str, method: str, api_format: str = None) -> OpenAPISchema:
        return self.get_response_schema(path, method, api_format=api_format)

    def _get_action_code_description(self, path: str, method: str) -> CodeDescriptionTuple:
        return self.get_action_code_description(path, method)

    def _get_responses(self, path: str, method: str) -> OpenAPISchema:
        return self.get_responses(path, method)

    def _map_field(self, field: Field) -> typing.Optional[OpenAPISchema]:
        return self.field_schema.map_field(field)

    def _get_field_description(self, field: Field) -> typing.Optional[str]:
        return self.field_schema.get_field_description(field)

    def _get_field_schema(self, field: Field) -> OpenAPISchema:
        return self.field_schema.get_field_schema(field)

    def _get_serializer_schema(
        self,
        serializer: BaseSerializer,
        write_only: bool = True,
        read_only: bool = True,
        required: bool = True,
    ) -> OpenAPISchema:
        return self.serializer_schema.get_serializer_schema(
            serializer, write_only=write_only, read_only=read_only, required=required
        )

    def _map_serializer(
        self,
        serializer: BaseSerializer,
        write_only: bool = True,
        read_only: bool = True,
        required: bool = True,
    ) -> OpenAPISchema:
        return self.serializer_schema.map_serializer(
            serializer, write_only=write_only, read_only=read_only, required=required
        )

    def _map_query_serializer(self, serializer: BaseSerializer) -> typing.List[OpenAPISchema]:
        return self.serializer_schema.map_query_serializer(serializer)



from __future__ import annotations
import typing

from restdoctor.utils.serializers import get_serializer_class_from_map

if typing.TYPE_CHECKING:
    from restdoctor.utils.serializers import SerializerClassMap, SerializerType
    from restdoctor.rest_framework.schema.custom_types import OpenAPISchema, ViewSchemaProtocol


class SerializerClassPaginationMixin:
    view_schema: typing.Optional[ViewSchemaProtocol]
    serializer_class: SerializerType
    serializer_class_map: SerializerClassMap

    def __init__(self, *args: typing.Any, view_schema: ViewSchemaProtocol = None, **kwargs: typing.Any):
        self.view_schema = view_schema
        super().__init__(*args, **kwargs)  # type: ignore

    def get_request_serializer_class(self) -> SerializerType:
        return get_serializer_class_from_map('pagination', 'request', self.serializer_class_map)

    def get_response_serializer_class(self) -> SerializerType:
        return get_serializer_class_from_map('pagination', 'response', self.serializer_class_map)

    def get_paginated_response_schema(self, schema: OpenAPISchema) -> OpenAPISchema:
        if self.view_schema:
            serializer_class = self.get_response_serializer_class()
            response_serializer = serializer_class()
            schema['properties']['meta'] = self.view_schema.get_serializer_schema(response_serializer)
        return schema



import pytest

from pytest_factoryboy import register
from restdoctor.rest_framework.pagination import (
    PageNumberPagination, PageNumberUncountedPagination,
    CursorUUIDPagination, CursorUUIDUncountedPagination,
)
from restdoctor.rest_framework.resources import ResourceViewSet, ResourceView, ResourceBase
from restdoctor.utils.api_prefix import get_api_path_prefixes
from tests.factories import MyModelFactory
from tests.stubs.models import MyModel


register(MyModelFactory)


@pytest.fixture()
def n_models(my_model_factory):
    def generate_n_models(n_models, **kwargs):
        return my_model_factory.create_batch(n_models, **kwargs)
    return generate_n_models


@pytest.fixture()
def my_models_queryset():
    return MyModel.objects.all()


@pytest.fixture()
def page_number_pagination():
    return PageNumberPagination()


@pytest.fixture()
def page_number_uncounted_pagination():
    return PageNumberUncountedPagination()


@pytest.fixture()
def cursor_uuid_pagination():
    return CursorUUIDPagination()


@pytest.fixture()
def cursor_uuid_uncounted_pagination():
    return CursorUUIDUncountedPagination()


@pytest.fixture()
def api_prefix() -> str:
    api_prefixes = get_api_path_prefixes()
    if api_prefixes:
        return api_prefixes[0]
    return ''


@pytest.fixture()
def resource_class(settings):
    def with_args(resource_discriminator, resource, base=None):

        class Base(base or ResourceViewSet):
            default_discriminative_value = settings.API_RESOURCE_DEFAULT
            resource_discriminative_param = settings.API_RESOURCE_DISCRIMINATIVE_PARAM

        resource_class = type(
            'TestResource', (Base,),
            {'resource_views_map': {resource_discriminator: resource}},
        )
        return resource_class
    return with_args


@pytest.fixture()
def resource_viewset_dispatch(mocker, resource_class):
    def with_args(resource_discriminator, resource_viewset, actions=None):
        viewset_class = resource_class(resource_discriminator, resource_viewset, base=ResourceViewSet)
        mocked_dispatch = mocker.patch.object(resource_viewset, 'dispatch')
        view_func = viewset_class.as_view(actions=actions)

        return view_func, mocked_dispatch

    return with_args


@pytest.fixture()
def resource_view_dispatch(mocker, resource_class):
    def with_args(resource_discriminator, resource_view):
        view_class = resource_class(resource_discriminator, resource_view, base=ResourceView)
        mocked_dispatch = mocker.patch.object(resource_view, 'dispatch')
        view_func = view_class.as_view()

        return view_func, mocked_dispatch

    return with_args


@pytest.fixture()
def get_discriminant_spy(mocker):
    return mocker.spy(ResourceBase, 'get_discriminant')



from __future__ import annotations
import datetime
from typing import Optional, Union, Any

from django.db import models
from rest_framework import ISO_8601
from rest_framework.fields import DateTimeField as BaseDateTimeField
from rest_framework.relations import HyperlinkedIdentityField as BaseHyperlinkedIdentityField
from rest_framework.request import Request
from rest_framework.settings import api_settings

from restdoctor.rest_framework.reverse import preserve_resource_params


class DateTimeField(BaseDateTimeField):
    def to_representation(
        self, value: datetime.datetime,
    ) -> Union[Optional[str], datetime.datetime]:
        if not value:
            return None

        output_format = getattr(self, 'format', api_settings.DATETIME_FORMAT)

        if output_format is None or isinstance(value, str):
            return value

        value = self.enforce_timezone(value)

        if output_format.lower() == ISO_8601:
            value = value.isoformat(timespec='microseconds')
            return value
        return value.strftime(output_format)


class HyperlinkedIdentityField(BaseHyperlinkedIdentityField):
    def get_url(self, obj: models.Model, view_name: str, request: Request, *args: Any, **kwargs: Any) -> str:
        url = super().get_url(obj, view_name, request, *args, **kwargs)
        return preserve_resource_params(url, request)



from __future__ import annotations

from rest_framework.decorators import action as drf_action
from rest_framework.fields import CharField, UUIDField
from rest_framework.serializers import Serializer

from restdoctor.rest_framework.resources import ResourceViewSet
from restdoctor.rest_framework.views import SerializerClassMapApiView
from restdoctor.rest_framework.viewsets import ListModelViewSet, ModelViewSet


class DefaultSerializer(Serializer):
    default_field = CharField()


class AnotherSerializer(Serializer):
    another_field = CharField()


class CreateSerializer(Serializer):
    create_field = CharField()


class ListSerializer(Serializer):
    list_field = CharField()


class ListFilterSerialiser(Serializer):
    filter_uuid_field = UUIDField(required=False)
    filter_field = CharField(allow_null=True, help_text='Help text')


class SerializerClassMapViewSet(ModelViewSet):
    serializer_class_map = {
        'default': DefaultSerializer,
        'list': {'response': ListSerializer},
        'create': {'response': CreateSerializer},
    }


class SerializerClassMapView(SerializerClassMapApiView):
    serializer_class_map = {
        'default': DefaultSerializer,
        'get': {'response': ListSerializer},
        'post': {'response': CreateSerializer},
    }

    def get(self, request, *args, **kwargs):
        pass

    def post(self, request, *args, **kwargs):
        pass


class DefaultViewSet(ModelViewSet):
    serializer_class = DefaultSerializer


class AnotherViewSet(ModelViewSet):
    serializer_class = AnotherSerializer


class ActionsViewSet(ModelViewSet):
    serializer_class = DefaultSerializer

    schema_action_codes_map = {'custom_action': {'200': None, '201': 'Description for 201 code.'}}

    @drf_action(detail=True, methods=['put'])
    def custom_action(self, *args, **kwargs):
        return {}


class ViewSetWithTags(ModelViewSet):
    schema_tags = ['tag1', 'tag2']
    serializer_class = DefaultSerializer


class ListViewSetWithRequestSerializer(ListModelViewSet):
    pagination_class = None
    schema_tags = ['tag']
    serializer_class_map = {'default': DefaultSerializer, 'list': {'request': ListFilterSerialiser}}


class ListViewSetWithoutRequestSerializer(ListModelViewSet):
    pagination_class = None
    serializer_class_map = {'default': DefaultSerializer}


class DefaultAnotherResourceViewSet(ResourceViewSet):
    default_discriminative_value = 'default'
    resource_views_map = {
        'default': DefaultViewSet,
        'another': AnotherViewSet,
        'actions': ActionsViewSet,
    }


class SingleResourceViewSet(ResourceViewSet):
    default_discriminative_value = 'default'
    resource_views_map = {'another': AnotherViewSet}



from __future__ import annotations

import json

from rest_framework.status import HTTP_200_OK, HTTP_201_CREATED, HTTP_202_ACCEPTED, HTTP_204_NO_CONTENT
from typing import Any, List, TYPE_CHECKING

from rest_framework.test import APIClient

if TYPE_CHECKING:
    from rest_framework.response import Response


class DRFClient(APIClient):
    def __init__(  # noqa CFQ002
        self, user: Any = None, authorization: str = '', accept: str = 'application/json',
        content_type: str = 'application/json', *args: Any, **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)

        self.user = user
        self.content_type = content_type
        self.accept = accept
        self.authorization = authorization

    def delete(self, *args: Any, **kwargs: Any) -> Response:
        return self._api_call('delete', *args, **kwargs)

    def get(self, *args: Any, **kwargs: Any) -> Response:
        return self._api_call('get', *args, **kwargs)

    def post(self, *args: Any, **kwargs: Any) -> Response:
        return self._api_call('post', *args, **kwargs)

    def put(self, *args: Any, **kwargs: Any) -> Response:
        return self._api_call('put', *args, **kwargs)

    def patch(self, *args: Any, **kwargs: Any) -> Response:
        return self._api_call('patch', *args, **kwargs)

    def delete_json(self, *args: Any, **kwargs: Any) -> Any:
        expected_status_codes = kwargs.pop(
            'expected_status_codes', [HTTP_200_OK, HTTP_202_ACCEPTED, HTTP_204_NO_CONTENT])
        return self._json_api_call('delete', *args, expected_status_codes=expected_status_codes, **kwargs)

    def get_json(self, *args: Any, **kwargs: Any) -> Any:
        expected_status_codes = kwargs.pop('expected_status_codes', [HTTP_200_OK])
        return self._json_api_call('get', *args, expected_status_codes=expected_status_codes, **kwargs)

    def post_json(self, *args: Any, **kwargs: Any) -> Any:
        expected_status_codes = kwargs.pop('expected_status_codes', [HTTP_200_OK, HTTP_201_CREATED])
        return self._json_api_call('post', *args, expected_status_codes=expected_status_codes, **kwargs)

    def put_json(self, *args: Any, **kwargs: Any) -> Any:
        expected_status_codes = kwargs.pop('expected_status_codes', [HTTP_200_OK])
        return self._json_api_call('put', *args, expected_status_codes=expected_status_codes, **kwargs)

    def patch_json(self, *args: Any, **kwargs: Any) -> Any:
        expected_status_codes = kwargs.pop('expected_status_codes', [HTTP_200_OK])
        return self._json_api_call('patch', *args, expected_status_codes=expected_status_codes, **kwargs)

    def _api_call(
        self, method: str, *args: Any, expected_status_codes: List[int] = None, **kwargs: Any,
    ) -> Response:
        accept = kwargs.get('accept') or kwargs.get('HTTP_ACCEPT') or self.accept
        authorization = kwargs.get('authorization') or kwargs.get('HTTP_AUTHORIZATION') or self.authorization
        content_type = kwargs.pop('content_type', self.content_type)

        if 'json' in kwargs:
            kwargs['data'] = json.dumps(kwargs.pop('json'))

        kwargs['HTTP_ACCEPT'] = accept
        kwargs['HTTP_AUTHORIZATION'] = authorization

        method = getattr(super(), method)
        response = method(*args, **kwargs, content_type=content_type)

        if expected_status_codes and response.status_code not in expected_status_codes:
            raise AssertionError(f'Response status code not in {expected_status_codes}')

        return response

    def _json_api_call(
        self, method: str, *args: Any, **kwargs: Any,
    ) -> Any:
        response = self._api_call(method, *args, **kwargs)
        return response.json()



from __future__ import annotations
import typing

from rest_framework.renderers import JSONRenderer

if typing.TYPE_CHECKING:
    from restdoctor.utils.custom_types import GenericContext
    from restdoctor.utils.media_type import APIParams


class RestDoctorRenderer(JSONRenderer):
    def __init__(self, media_type: str, api_params: typing.Optional[APIParams]) -> None:
        self.media_type = media_type
        self.api_params = api_params

    def render(
        self,
        data: GenericContext,
        accepted_media_type: str = None,
        renderer_context: typing.Optional[GenericContext] = None,
    ) -> bytes:
        renderer_context: GenericContext = renderer_context or {}
        if data is None or 'message' in data:
            result = data
        else:
            result = {'data': data}
            meta = renderer_context.get('meta')
            if meta:
                result['meta'] = meta

        verbose = bool(self.api_params and self.api_params.format.endswith('verbose'))
        if result is not None and verbose:
            result['query'] = {
                'args': renderer_context.get('args', []),
                **renderer_context.get('kwargs', {}),
            }

        return super().render(result, accepted_media_type, renderer_context)



import datetime

import pytest
import pytz
from django.utils.timezone import make_aware

from restdoctor.rest_framework.fields import DateTimeField


@pytest.mark.parametrize(
    'datetime_obj, expected_string_representation',
    (
        (datetime.datetime(2020, 7, 16), '2020-07-16T00:00:00.000000+03:00'),
        (datetime.datetime(2020, 7, 16, tzinfo=pytz.utc), '2020-07-16T03:00:00.000000+03:00'),
        (make_aware(datetime.datetime(2020, 7, 16)), '2020-07-16T00:00:00.000000+03:00'),
        (make_aware(datetime.datetime(2020, 7, 16, 18, 0, 0)), '2020-07-16T18:00:00.000000+03:00'),
        (make_aware(datetime.datetime(2020, 7, 16, 18, 0, 0, 48)), '2020-07-16T18:00:00.000048+03:00'),
    ),
)
def test_datetime_field_to_representation(
    datetime_obj, expected_string_representation,
):
    assert DateTimeField().to_representation(datetime_obj) == expected_string_representation



from __future__ import annotations

import contextlib
import copy
import inspect
import typing

from django.db.models import Model
from rest_framework.fields import SerializerMethodField
from rest_framework.serializers import Serializer, ListSerializer

from restdoctor.django.sensitive_data import get_model_sensitive_fields, is_model_field_sensitive

if typing.TYPE_CHECKING:
    SerializerClassOrInstance = typing.Union[typing.Type[Serializer], Serializer]
    SensitiveDataConfig = typing.Dict[str, typing.Any]
    SensitiveDataConfigValue = typing.Union[bool, SensitiveDataConfig]
    SerializerData = typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Dict[str, typing.Any]]]


def get_serializer_sensitive_fields(
    serializer: Serializer, with_model_sensitive_fields: bool = True,
) -> typing.List[str]:
    sensitive_fields = []

    with contextlib.suppress(AttributeError, TypeError):
        sensitive_fields.extend(serializer.SensitiveData.include)

    if with_model_sensitive_fields:
        with contextlib.suppress(AttributeError, TypeError):
            sensitive_fields.extend(get_model_sensitive_fields(serializer.Meta.model))

    return sensitive_fields


def get_serializer_instance(serializer: SerializerClassOrInstance) -> Serializer:
    return serializer() if inspect.isclass(serializer) else serializer


def get_field_serializer(
    serializer: Serializer, field_name: str, field: typing.Any,
) -> typing.Optional[Serializer]:
    if isinstance(field, Serializer):
        return field
    elif isinstance(field, ListSerializer):
        return field.child
    elif isinstance(field, SerializerMethodField):
        if (
            field_name in serializer.__class__._declared_fields
            and hasattr(serializer.__class__._declared_fields[field_name], 'schema_type')
            and isinstance(serializer.__class__._declared_fields[field_name].schema_type, Serializer)
        ):
            return serializer.__class__._declared_fields[field_name].schema_type


def get_serializer_sensitive_data_config(
    serializer: SerializerClassOrInstance,
) -> SensitiveDataConfig:
    sensitive_data_config: SensitiveDataConfig = {}

    serializer = get_serializer_instance(serializer)

    try:
        model = serializer.Meta.model
    except AttributeError:
        model = None

    for field_name, field in serializer.get_fields().items():
        result = get_field_sensitive_data_config(
            field_name, field, serializer, model,
        )
        if result:
            sensitive_data_config[field_name] = result

    return sensitive_data_config


def get_field_sensitive_data_config(
    field_name: str,
    field: typing.Any,
    serializer: Serializer,
    model: typing.Optional[Model],
) -> SensitiveDataConfigValue:
    result: SensitiveDataConfigValue = False

    field_serializer = get_field_serializer(serializer, field_name, field)
    field_source = getattr(field, 'source', None)

    if field_name in get_serializer_sensitive_fields(serializer):
        result = True
    elif field_serializer is not None:
        result = get_serializer_sensitive_data_config(field_serializer)
    elif field_source is not None and field_source != '*' and model:
        result = is_model_field_sensitive(model, field.source)

    return result


def clear_sensitive_data(
    data: SerializerData,
    serializer: SerializerClassOrInstance,
) -> SerializerData:
    data = copy.deepcopy(data)

    try:
        sensitive_data_config = get_serializer_sensitive_data_config(serializer)
    except AttributeError:
        return data

    if not sensitive_data_config:
        return data

    for field_name, field_value in sensitive_data_config.items():
        clear_sensitive_field(field_name, field_value, data)

    return data


def clear_sensitive_field(
    field_name: str, field_value: typing.Any, data: SerializerData,
) -> None:
    if isinstance(data, list):
        for data_item in data:
            clear_sensitive_field(field_name, field_value, data_item)
    elif isinstance(data, dict) and field_name in data:
        if isinstance(field_value, dict):
            for related_field_name, related_field_value in field_value.items():
                clear_sensitive_field(related_field_name, related_field_value, data[field_name])
        else:
            data[field_name] = '[Cleaned]'



import pytest

from restdoctor.utils.api_prefix import get_api_prefixes, get_api_path_prefixes


@pytest.mark.parametrize(
    'api_prefixes,default,expected_result',
    (
        ('/prefix', None, ('/prefix',)),
        (('/prefix',), None, ('/prefix',)),
        (None, None, ()),
        (None, '/prefix', ('/prefix',)),
        (('/prefix1', '/prefix2'), None, ('/prefix1', '/prefix2')),
    ),
)
def test_get_api_prefixes_success_case(api_prefixes, default, expected_result, settings):
    settings.API_PREFIXES = api_prefixes
    result = get_api_prefixes(default=default)

    assert result == expected_result


@pytest.mark.parametrize(
    'api_prefixes,default,expected_result',
    (
        ('/prefix', None, ('prefix/',)),
        (('/prefix',), None, ('prefix/',)),
        (None, None, ()),
        (None, '/prefix', ('prefix/',)),
        (('/prefix1', '/prefix2'), None, ('prefix1/', 'prefix2/')),
    ),
)
def test_get_api_path_prefixes_success_case(api_prefixes, default, expected_result, settings):
    settings.API_PREFIXES = api_prefixes
    result = get_api_path_prefixes(default=default)

    assert result == expected_result



from __future__ import annotations

from restdoctor.rest_framework.schema.generators import RefsSchemaGenerator
from restdoctor.rest_framework.schema.openapi import RestDoctorSchema
from restdoctor.rest_framework.schema.refs_registry import LocalRefsRegistry
from restdoctor.rest_framework.schema.resources import ResourceSchema
from restdoctor.rest_framework.schema.wrappers import SchemaWrapper

__all__ = [
    'RefsSchemaGenerator', 'RestDoctorSchema', 'LocalRefsRegistry', 'ResourceSchema', 'SchemaWrapper',
]



import pytest

from restdoctor.rest_framework.schema import RefsSchemaGenerator
from tests.test_unit.test_schema.stubs import DefaultViewSet


@pytest.mark.parametrize(
    'generator,expected_schema_keys',
    (
        (None, ['type', 'properties', 'required']),
        (RefsSchemaGenerator(), ['$ref']),
    ),
)
def test_schema_get_operation_with_refs_schema_generator_success_case(
    get_create_view_func, generator, expected_schema_keys,
):
    create_view = get_create_view_func('test', DefaultViewSet, 'test')

    view = create_view('/test/', 'GET')
    view.schema.generator = generator
    operation = view.schema.get_operation('/test/', 'GET')
    responses_schema = operation['responses']['200']['content']['application/vnd.vendor']['schema']
    responses_properties = responses_schema['properties']['data']
    if 'items' in responses_properties:
        responses_properties = responses_properties['items']

    assert list(responses_properties.keys()) == expected_schema_keys



from __future__ import annotations

import typing

from django.utils.encoding import force_str
from rest_framework.fields import HiddenField
from rest_framework.serializers import BaseSerializer

from restdoctor.rest_framework.schema.custom_types import (
    OpenAPISchema,
    SerializerSchemaProtocol,
    ViewSchemaProtocol,
)


class SerializerSchema(SerializerSchemaProtocol):
    def __init__(self, view_schema: ViewSchemaProtocol):
        self.view_schema = view_schema

    def map_serializer_fields(
        self,
        serializer: BaseSerializer,
        include_write_only: bool = True,
        include_read_only: bool = True,
    ) -> typing.Tuple[OpenAPISchema, typing.List[str]]:
        required_list = []
        properties = {}

        for field in serializer.fields.values():
            if (
                isinstance(field, HiddenField)
                or (field.write_only and not include_write_only)
                or (field.read_only and not include_read_only)
            ):
                continue
            if field.required:
                required_list.append(field.field_name)
            field_schema = self.view_schema.get_field_schema(field)
            if field_schema:
                properties[field.field_name] = field_schema
        return properties, required_list

    def get_serializer_schema(
        self,
        serializer: BaseSerializer,
        write_only: bool = True,
        read_only: bool = True,
        required: bool = True,
    ) -> OpenAPISchema:
        properties, required_list = self.map_serializer_fields(
            serializer, include_write_only=write_only, include_read_only=read_only
        )

        if not properties:
            return {}

        schema: OpenAPISchema = {'type': 'object', 'properties': properties}
        if serializer.__doc__:
            schema['description'] = serializer.__doc__
        if required and required_list:
            schema['required'] = required_list
        return schema

    def get_ref_name(
        self,
        serializer: BaseSerializer,
        write_only: bool = True,
        read_only: bool = True,
        required: bool = True,
    ) -> str:
        serializer_module_name = serializer.__module__.split('.')[0]
        serializer_class_name = serializer.__class__.__name__

        suffixes = []
        if write_only:
            suffixes.append('W')
        if read_only:
            suffixes.append('R')
        if required:
            suffixes.append('Q')

        suffix = ''.join(suffixes)

        if serializer_class_name.endswith('Serializer'):
            serializer_class_name = serializer_class_name[:-10]
        return f'#/components/schemas/{serializer_module_name}_{serializer_class_name}{suffix}'

    def map_serializer(
        self,
        serializer: BaseSerializer,
        write_only: bool = True,
        read_only: bool = True,
        required: bool = True,
    ) -> OpenAPISchema:
        schema = self.get_serializer_schema(
            serializer, write_only=write_only, read_only=read_only, required=required
        )

        if schema and self.view_schema.generator:
            ref = self.get_ref_name(
                serializer, write_only=write_only, read_only=read_only, required=required
            )
            self.view_schema.generator.local_refs_registry.put_local_ref(ref, schema)
            schema = {'$ref': ref}

        return schema

    def map_query_serializer(self, serializer: BaseSerializer) -> typing.List[OpenAPISchema]:
        props = []

        for field in serializer.fields.values():
            if isinstance(field, HiddenField):
                continue

            prop = {
                'name': field.field_name,
                'in': 'query',
                'required': field.required,
                'schema': self.view_schema.get_field_schema(field),
            }

            if field.help_text:
                prop['description'] = force_str(field.help_text)

            props.append(prop)

        return props



from __future__ import annotations
import typing

from rest_framework.negotiation import DefaultContentNegotiation

from restdoctor.rest_framework.parsers import BestDoctorParser
from restdoctor.utils.media_type import get_media_type
from restdoctor.rest_framework.renderers import RestDoctorRenderer

if typing.TYPE_CHECKING:
    from django.http import HttpRequest
    from restdoctor.rest_framework.custom_types import (
        Parsers, OptionalParser, OptionalRenderer, Renderers,
    )


class APIVersionContentNegotiation(DefaultContentNegotiation):
    def select_renderer(
        self, request: HttpRequest, renderers: Renderers, format_suffix: str,
    ) -> typing.Tuple[OptionalRenderer, str]:
        api_params = getattr(request, 'api_params', None)
        if api_params is not None:
            media_type = get_media_type(api_params)
            return RestDoctorRenderer(media_type, api_params), media_type

        return super().select_renderer(request, renderers, format_suffix)

    def select_parser(self, request: HttpRequest, parsers: Parsers) -> OptionalParser:
        api_params = getattr(request, 'api_params', None)
        parser = BestDoctorParser('application/json', api_params)

        return super().select_parser(request, [parser, *parsers])



from rest_framework.serializers import BaseSerializer


class SerializerA(BaseSerializer):
    pass


class SerializerB(BaseSerializer):
    pass


class SerializerC(BaseSerializer):
    pass


class SerializerD(BaseSerializer):
    pass


class SerializerE(BaseSerializer):
    pass


serializer_class_map_with_default = {
    'default': SerializerA,
    'retrieve': {
        'request': SerializerB,
    },
    'list': {
        'request': SerializerB,
        'response': SerializerB,
    },
    'create': {
        'request': SerializerC,
        'response': SerializerB,
    },
}

serializer_class_map_no_default = {
    'retrieve': {
        'request': SerializerB,
    },
    'list': {
        'request': SerializerB,
        'response': SerializerB,
    },
    'create': {
        'request': SerializerC,
        'response': SerializerB,
    },
}


serializer_class_map_with_format = {
    'default': SerializerA,
    'default.compact': SerializerB,
    'retrieve': {
        'response': SerializerB,
        'response.compact': SerializerC,
    },
    'list': {
        'response.full': SerializerE,
    },
}



import pytest

from restdoctor.utils.serializers import get_serializer_class_from_map

from tests.test_unit.test_serializers.test_serializer_class_map.stubs import (
    SerializerA, SerializerB, SerializerC, SerializerD, SerializerE,
    serializer_class_map_with_default, serializer_class_map_no_default,
    serializer_class_map_with_format,
)


@pytest.mark.parametrize(
    'serializer_class_map,action,stage,default,expected',
    (
        (serializer_class_map_with_default, 'retrieve', 'request', SerializerD, SerializerB),
        (serializer_class_map_with_default, 'retrieve', 'response', SerializerD, SerializerA),
        (serializer_class_map_with_default, 'list', 'response', SerializerD, SerializerB),
        (serializer_class_map_no_default, 'retrieve', 'request', SerializerD, SerializerB),
        (serializer_class_map_no_default, 'retrieve', 'response', SerializerD, SerializerD),
        (serializer_class_map_no_default, 'list', 'response', SerializerD, SerializerB),

    ),
)
def test_get_serializer_class_from_map(serializer_class_map, action, stage, default, expected):
    serializer_class = get_serializer_class_from_map(action, stage, serializer_class_map, default)

    assert serializer_class == expected


@pytest.mark.parametrize(
    'serializer_class_map,action,stage,api_format,default,expected',
    (
        (serializer_class_map_with_format, 'retrieve', 'request', None, SerializerD, SerializerB),
        (serializer_class_map_with_format, 'retrieve', 'response', None, SerializerD, SerializerC),
        (serializer_class_map_with_format, 'retrieve', 'response', 'full', SerializerD, SerializerB),
        (serializer_class_map_with_format, 'list', 'response', 'with_related', SerializerD, SerializerA),
        (serializer_class_map_with_format, 'list', 'request', None, SerializerD, SerializerB),
        (serializer_class_map_with_format, 'list', 'response', 'full', SerializerD, SerializerE),
    ),
)
def test_get_serializer_class_from_map_with_format(
    serializer_class_map, action, stage, api_format, default, expected, settings,
):
    settings.API_DEFAULT_FORMAT = 'compact'

    serializer_class = get_serializer_class_from_map(
        action, stage, serializer_class_map, default, api_format=api_format)

    assert serializer_class == expected



from __future__ import annotations

import collections
import contextlib
import decimal
import typing

from django.conf import settings
from django.contrib.gis.db.models import PointField
from django.core.exceptions import FieldDoesNotExist, ImproperlyConfigured
from django.core.validators import RegexValidator
from django.db.models import AutoField
from rest_framework.fields import (
    BooleanField,
    ChoiceField,
    DateField,
    DateTimeField,
    DecimalField,
    DictField,
    EmailField,
    Field,
    FileField,
    FloatField,
    HStoreField,
    IntegerField,
    IPAddressField,
    JSONField,
    ListField,
    ModelField,
    MultipleChoiceField,
    SerializerMethodField,
    URLField,
    UUIDField,
    _UnvalidatedField,
    empty,
)
from rest_framework.relations import ManyRelatedField, PrimaryKeyRelatedField
from rest_framework.schemas.openapi import AutoSchema
from rest_framework.serializers import BaseSerializer, ListSerializer, ModelSerializer, Serializer
from rest_framework.settings import api_settings

from restdoctor.rest_framework.schema.custom_types import (
    FieldSchemaProtocol,
    OpenAPISchema,
    ViewSchemaProtocol,
)


def array_schema(item_schema: OpenAPISchema = None) -> OpenAPISchema:
    schema: OpenAPISchema = {'type': 'array'}
    if item_schema:
        schema['items'] = item_schema
    return schema


def string_schema(string_format: str = None) -> OpenAPISchema:
    schema: OpenAPISchema = {'type': 'string'}
    if string_format:
        schema['format'] = string_format
    return schema


def drf_map_field_validators(obj: FieldSchemaProtocol, field: Field, schema: OpenAPISchema) -> None:
    try:
        AutoSchema.map_field_validators(obj, field, schema)
    except AttributeError:
        AutoSchema._map_field_validators(obj, field, schema)


class FieldSchema(FieldSchemaProtocol):
    def __init__(self, view_schema: ViewSchemaProtocol):
        self.view_schema = view_schema

    def get_field_schema(self, field: Field) -> OpenAPISchema:
        schema = self.map_field(field)
        if not schema:
            return {}
        if field.read_only:
            schema['readOnly'] = True
        if field.write_only:
            schema['writeOnly'] = True
        if field.allow_null:
            schema['nullable'] = True
        if field.default and field.default != empty and not callable(field.default):
            schema['default'] = field.default
        description = self.get_field_description(field)
        if description:
            schema['description'] = description

        self.map_field_validators(field, schema)
        return schema

    def get_field_description(self, field: Field) -> typing.Optional[str]:
        field_description = None
        if field.help_text:
            field_description = str(field.help_text)
        elif isinstance(field.parent, ModelSerializer):
            if isinstance(field, ModelField):
                field_description = field.model_field.verbose_name
            elif field.source != '*':
                with contextlib.suppress(AttributeError, LookupError, FieldDoesNotExist):
                    field_description = str(
                        field.parent.Meta.model._meta.get_field(field.source).verbose_name
                    )

        if settings.API_STRICT_SCHEMA_VALIDATION and not field_description:
            raise ImproperlyConfigured(
                f'field {field.field_name} in serializer {field.parent.__class__.__name__} '
                f'should have help_text argument or verbose_name in source model field'
            )

        return field_description

    def map_field_validators(self, field: Field, schema: OpenAPISchema) -> None:
        drf_map_field_validators(self, field, schema)

        # Backported from django-rest-framework
        # https://github.com/encode/django-rest-framework/commit/5ce237e00471d885f05e6d979ec777552809b3b1
        for validator in field.validators:
            if isinstance(validator, RegexValidator):
                schema['pattern'] = validator.regex.pattern.replace('\\Z', '\\z')

    def map_field(self, field: Field) -> typing.Optional[OpenAPISchema]:
        # Field.__deepcopy__ сбрасывает кастомные атрибуты, поэтому схему ищем через класс сериализатора
        if isinstance(field, SerializerMethodField):
            parent_field = field.parent.__class__._declared_fields[field.field_name]
            field = getattr(parent_field, 'schema_type', field)

        field_handlers = [
            (_UnvalidatedField, lambda _: None),
            (PointField, self.map_point_field),
            ((ListSerializer, Serializer), self.map_serializer),
            ((ManyRelatedField, PrimaryKeyRelatedField), self.map_related),
            (MultipleChoiceField, self.map_choice_field),
            (ChoiceField, self.map_choice_field),
            (ListField, lambda f: array_schema(self.map_field(f.child))),
            (DateField, lambda _: string_schema('date')),
            (DateTimeField, lambda _: string_schema('date-time')),
            (EmailField, lambda _: string_schema('email')),
            (URLField, lambda _: string_schema('uri')),
            (UUIDField, lambda _: string_schema('uuid')),
            (IPAddressField, self.map_ipaddress_field),
            (DecimalField, self.map_decimal_field),
            ((FloatField, IntegerField), self.map_numeric_field),
            (FileField, lambda _: string_schema('binary')),
            (BooleanField, lambda _: {'type': 'boolean'}),
            ((JSONField, DictField, HStoreField), lambda _: {'type': 'object'}),
            (Field, lambda _: string_schema()),
            # Catch-all type
            (object, lambda _: string_schema()),
        ]

        for field_types, handler in field_handlers:
            if isinstance(field, field_types):
                return handler(field)

    def map_choice_field(self, field: ChoiceField) -> OpenAPISchema:
        choices = list(
            collections.OrderedDict.fromkeys(field.choices)
        )  # preserve order and remove duplicates
        schema: OpenAPISchema = {'enum': choices}

        choice_type = None
        if all(isinstance(choice, bool) for choice in choices):
            choice_type = 'boolean'
        elif all(isinstance(choice, int) for choice in choices):
            choice_type = 'integer'
        elif all(isinstance(choice, (int, float, decimal.Decimal)) for choice in choices):
            choice_type = 'number'
        elif all(isinstance(choice, str) for choice in choices):
            choice_type = 'string'

        if choice_type:
            schema['type'] = choice_type

        if isinstance(field, MultipleChoiceField):
            return array_schema(schema)
        return schema

    def map_point_field(self, field: PointField) -> OpenAPISchema:
        return {
            'type': {'type': 'Point'},
            'coordinates': {
                'type': 'array',
                'items': {'type': 'number', 'format': 'float'},
                'example': [12.9721, 77.5933],
                'minItems': 2,
                'maxItems': 3,
            },
        }

    def map_serializer(self, field: BaseSerializer) -> OpenAPISchema:
        if isinstance(field, ListSerializer):
            return array_schema(self.view_schema.map_serializer(field.child))
        else:
            schema = self.view_schema.map_serializer(field)
            schema['type'] = 'object'
            return schema

    def map_related(
        self, field: typing.Union[ManyRelatedField, PrimaryKeyRelatedField]
    ) -> typing.Optional[OpenAPISchema]:
        if isinstance(field, ManyRelatedField):
            return array_schema(self.map_field(field.child_relation))
        if isinstance(field, PrimaryKeyRelatedField):
            model = getattr(field.queryset, 'model', None)
            if model is not None:
                model_field = model._meta.pk
                if isinstance(model_field, AutoField):
                    return {'type': 'integer'}

    def map_list_field(self, field: ListField) -> OpenAPISchema:
        return array_schema(self.map_field(field.child))

    def map_decimal_field(self, field: DecimalField) -> OpenAPISchema:
        if getattr(field, 'coerce_to_string', api_settings.COERCE_DECIMAL_TO_STRING):
            schema = string_schema('decimal')
        else:
            schema = {'type': 'number'}

        if field.decimal_places:
            schema['multipleOf'] = float('.' + (field.decimal_places - 1) * '0' + '1')
        if field.max_whole_digits:
            value = int(field.max_whole_digits * '9') + 1
            schema['maximum'] = value
            schema['minimum'] = -value

        if field.max_value:
            schema['maximum'] = field.max_value
        if field.min_value:
            schema['minimum'] = field.min_value

        return schema

    def map_numeric_field(self, field: Field) -> OpenAPISchema:
        schema = {'type': 'number'}
        if field.max_value:
            schema['maximum'] = field.max_value
        if field.min_value:
            schema['minimum'] = field.min_value

        if isinstance(field, IntegerField):
            schema = {'type': 'integer'}
            max_value = int(field.max_value or 0)
            min_value = int(field.min_value or 0)
            if max(max_value, min_value) > 2147483647:
                schema['format'] = 'int64'

        return schema

    def map_ipaddress_field(self, field: IPAddressField) -> OpenAPISchema:
        string_format = field.protocol if field.protocol != 'both' else None
        return string_schema(string_format)



from __future__ import annotations

import datetime
import typing

from django.conf import settings
from django.contrib.auth import get_user_model
from jwt import decode as jwt_decode, encode as jwt_encode

if typing.TYPE_CHECKING:
    from django.http import HttpRequest
    from django.contrib.auth.models import User


def decode_token(token: str) -> typing.Mapping:
    return jwt_decode(
        token,
        settings.SECRET_KEY,
        settings.JWT_VERIFY,
        options={
            'verify_exp': settings.JWT_VERIFY_EXPIRATION,
        },
        leeway=settings.JWT_LEEWAY,
        audience=settings.JWT_AUDIENCE,
        issuer=settings.JWT_ISSUER,
        algorithms=[settings.JWT_ALGORITHM],
    )


def encode_token(payload: typing.Dict) -> bytes:
    return jwt_encode(
        payload,
        settings.SECRET_KEY,
        algorithm=settings.JWT_ALGORITHM,
    )


def get_username_from_request_token(request: HttpRequest) -> typing.Optional[str]:
    header_parts = request.META.get('HTTP_AUTHORIZATION', '').split()
    if len(header_parts) == 2:
        if header_parts[0] == 'Bearer':
            payload = decode_token(header_parts[-1])

            return payload.get(get_user_model().USERNAME_FIELD)


def get_payload(user: User) -> typing.Dict:
    username = user.get_username()

    if hasattr(username, 'pk'):
        username = username.pk

    payload = {
        user.USERNAME_FIELD: username,
        'exp': datetime.datetime.utcnow() + settings.JWT_EXPIRATION_WEB_DELTA,
    }

    return payload


def get_token(user: User) -> bytes:
    payload = get_payload(user)
    return encode_token(payload)



import pytest


@pytest.mark.parametrize(
    'endpoint, accept, expected_content_type',
    [
        (
            'empty_v1',
            'application/vnd.restdoctor',
            'application/vnd.restdoctor.v1.full+json',
        ),
        (
            'empty_v1',
            'application/vnd.restdoctor.v1',
            'application/vnd.restdoctor.v1.full+json',
        ),
        (
            'empty_v1',
            'application/vnd.restdoctor.v1.compact',
            'application/vnd.restdoctor.v1.compact+json',
        ),
        (
            'empty_v1',
            'application/vnd.restdoctor.v1.compact-verbose',
            'application/vnd.restdoctor.v1.compact-verbose+json',
        ),
        (
            'empty_v2',
            'application/vnd.restdoctor.v2',
            'application/vnd.restdoctor.v2.full+json',
        ),
        (
            'empty_v2',
            'application/vnd.restdoctor.v2.compact',
            'application/vnd.restdoctor.v2.compact+json',
        ),
        (
            'empty_v2',
            'application/vnd.restdoctor.v2.compact-verbose',
            'application/vnd.restdoctor.v2.compact-verbose+json',
        ),
        (
            'empty_v2',
            'application/vnd.restdoctor.v2.wrong-format',
            'application/vnd.restdoctor.v2.full+json',
        ),
    ],
)
@pytest.mark.django_db
def test_accept_header_version_success_case(
    client, api_prefix, endpoint, accept, expected_content_type, settings,
):
    settings.API_VERSIONS = {
        'v1': 'tests.stubs.api.v1_urls',
        'v2': 'tests.stubs.api.v2_urls',
    }
    settings.API_FORMATS = ('full', 'compact', 'compact-verbose')
    settings.API_VENDOR_STRING = 'RestDoctor'

    response = client.get(f'/{api_prefix}{endpoint}', HTTP_ACCEPT=accept)

    assert response.status_code == 200
    assert response['Content-Type'] == expected_content_type


@pytest.mark.parametrize(
    'endpoint, accept',
    [
        (
            'empty_v1',
            'application/json',
        ),
    ],
)
@pytest.mark.django_db
def test_accept_header_fallback_success_case(client, api_prefix, endpoint, accept, settings):
    settings.API_VERSIONS = {
        'v1': 'tests.stubs.api.v1_urls',
        'v2': 'tests.stubs.api.v2_urls',
    }

    response = client.get(f'/{api_prefix}{endpoint}', HTTP_ACCEPT=accept)

    assert response.status_code == 404


@pytest.mark.parametrize(
    'endpoint, content_type',
    [
        (
            'empty_v1',
            'application/vnd.restdoctor.v1.full+json',
        ),
    ],
)
@pytest.mark.django_db
def test_post_without_accept_header_success_case(client, api_prefix, endpoint, content_type, settings):
    settings.API_VERSIONS = {
        'v1': 'tests.stubs.api.v1_urls',
        'v2': 'tests.stubs.api.v2_urls',
    }
    settings.API_VENDOR_STRING = 'RestDoctor'

    response = client.post(
        f'/{api_prefix}{endpoint}', data='{}',
        content_type='application/json', HTTP_ACCEPT='application/vnd.restdoctor',
    )

    assert response.status_code == 200
    assert response['Content-Type'] == content_type



from __future__ import annotations

import re
import contextlib
from typing import Optional, Dict

from django.core.exceptions import (
    ObjectDoesNotExist, ValidationError as DjangoValidationError, ImproperlyConfigured,
)
from django.db.models import Model
from django.http import Http404
from rest_framework.generics import GenericAPIView as BaseGenericAPIView

from restdoctor.rest_framework.mixins import NegotiatedMixin


class GenericAPIView(NegotiatedMixin, BaseGenericAPIView):
    lookup_fields: Optional[Dict[str, str]] = None

    def get_object(self) -> Model:
        if self.lookup_fields is None:
            return super().get_object()

        self._check_lookup_configuration()
        queryset = self.filter_queryset(self.get_queryset())

        obj = None
        filter_value = self.kwargs[self.lookup_url_kwarg]
        for lookup_field, lookup_regex in self.lookup_fields.items():
            if re.match(lookup_regex, filter_value):
                filter_kwargs = {lookup_field: filter_value}
                with contextlib.suppress(
                    AttributeError, TypeError, ValueError, DjangoValidationError, ObjectDoesNotExist,
                ):
                    obj = queryset.get(**filter_kwargs)

        if obj is None:
            raise Http404

        self.check_object_permissions(self.request, obj)

        return obj

    def _check_lookup_configuration(self) -> None:
        if not self.lookup_url_kwarg:
            raise ImproperlyConfigured((
                f'View {self.__class__.__name__} improperly configured. '
                f'Set the `.lookup_url_kwarg` with `.lookup_fields`'
            ))

        if self.lookup_url_kwarg not in self.kwargs:
            raise ImproperlyConfigured((
                f'Expected view {self.__class__.__name__} to be called with a URL keyword argument '
                f'named "{self.lookup_url_kwarg}". Fix your URL conf, or set the `.lookup_field` '
                f'attribute on the view correctly.'
            ))



import pytest

from restdoctor.rest_framework.test_client import DRFClient


@pytest.fixture()
def drf_client():
    return DRFClient()


@pytest.fixture()
def mock_api_call_factory(mocker):
    def mock_api_call(method):
        return mocker.patch('restdoctor.rest_framework.test_client.APIClient.' + method)
    return mock_api_call



from restdoctor.rest_framework.pagination.page_number import (  # noqa: F401
    PageNumberPagination, PageNumberUncountedPagination,
)
from restdoctor.rest_framework.pagination.cursor_uuid import (  # noqa: F401
    CursorUUIDPagination, CursorUUIDUncountedPagination, get_order,
)



from __future__ import annotations

import pytest

from restdoctor.rest_framework.routers import ResourceRouter
from restdoctor.rest_framework.schema import ResourceSchema
from tests.test_unit.test_schema.stubs import DefaultAnotherResourceViewSet, SingleResourceViewSet


@pytest.mark.parametrize(
    ('view_class_name', 'expected_object_name'),
    [
        ('PatientView', 'Patient'),
        ('PatientAPIView', 'Patient'),
        ('PatientViewAPIView', 'PatientView'),
        ('PatientResourceViewSet', 'PatientResource'),
    ],
)
def test_object_name_success_case(view_class_name, expected_object_name):
    schema = ResourceSchema()
    view_class = type(view_class_name, (), {})
    schema.view = view_class()

    object_name = schema._get_object_name('/test/', 'GET', 'action')

    assert object_name == expected_object_name


def test_get_response_schema_success_case(get_create_view_func):
    create_view = get_create_view_func(
        'test', DefaultAnotherResourceViewSet, 'test', router=ResourceRouter()
    )

    view = create_view('/test/', 'GET')
    response_schema = view.schema.get_resources_response_schema('/test/', 'GET')

    assert len(response_schema.get('oneOf', [])) == 2


def test_get_response_schema_wrong_method_success_case(get_create_view_func):
    create_view = get_create_view_func(
        'test', DefaultAnotherResourceViewSet, 'test', router=ResourceRouter()
    )

    view = create_view('/test/', 'POST')
    response_schema = view.schema.get_resources_response_schema('/test/', 'POST')

    assert len(response_schema.get('oneOf', [])) == 0


def test_get_request_body_schema_success_case(get_create_view_func):
    create_view = get_create_view_func(
        'test', DefaultAnotherResourceViewSet, 'test', router=ResourceRouter()
    )

    view = create_view('/test/', 'POST')
    view.resource_discriminate_methods = ['POST']
    request_body = view.schema.get_resources_request_body_schema('/test/', 'POST')

    assert len(request_body.get('oneOf', [])) == 2


def test_get_responses_success_case(
    get_create_view_func, resource_default_rq_schema, resource_another_rq_schema
):
    create_view = get_create_view_func(
        'test', DefaultAnotherResourceViewSet, 'test', router=ResourceRouter()
    )

    view = create_view('/test/', 'POST')
    view.resource_discriminate_methods = ['POST']
    responses = view.schema.get_responses('/test/', 'POST')
    content = responses['201']['content']

    assert content['application/vnd.vendor']['schema']['oneOf'] == [
        resource_default_rq_schema,
        resource_another_rq_schema,
    ]
    assert content['application/vnd.vendor.v1-default']['schema'] == resource_default_rq_schema
    assert content['application/vnd.vendor.v1-another']['schema'] == resource_another_rq_schema
    assert content['application/vnd.vendor.v1-actions']['schema'] == resource_default_rq_schema


def test_get_request_body_success_case(
    get_create_view_func, resource_default_ref, resource_another_ref
):
    create_view = get_create_view_func(
        'test', DefaultAnotherResourceViewSet, 'test', router=ResourceRouter()
    )
    resource_default_wq_ref = resource_default_ref('WQ')
    resource_another_wq_ref = resource_another_ref('WQ')

    view = create_view('/test/', 'POST')
    view.resource_discriminate_methods = ['POST']
    request_body = view.schema.get_request_body('/test/', 'POST')
    content = request_body['content']

    assert content['application/vnd.vendor']['schema']['oneOf'] == [
        resource_default_wq_ref,
        resource_another_wq_ref,
    ]
    assert content['application/vnd.vendor.v1-default']['schema'] == resource_default_wq_ref
    assert content['application/vnd.vendor.v1-another']['schema'] == resource_another_wq_ref
    assert content['application/vnd.vendor.v1-actions']['schema'] == resource_default_wq_ref


def test_get_action_responses_success_case(get_create_view_func, resource_default_rq_schema):
    create_view = get_create_view_func(
        'test', DefaultAnotherResourceViewSet, 'test', router=ResourceRouter()
    )

    view = create_view('/test/123/custom_action/', 'PUT')
    view.resource_discriminate_methods = ['PUT']
    responses = view.schema.get_responses('/test/123/custom_action/', 'PUT')
    content = responses['201']['content']

    assert content['application/vnd.vendor']['schema'] == resource_default_rq_schema
    assert 'application/vnd.vendor.v1-default' not in content
    assert 'application/vnd.vendor.v1-another' not in content
    assert content['application/vnd.vendor.v1-actions']['schema'] == resource_default_rq_schema


@pytest.mark.parametrize(
    ('viewset_class'),
    [(DefaultAnotherResourceViewSet), (SingleResourceViewSet)],
    ids=[
        'get params from defauilt viewset in resource',
        'get params from single viewset in resource',
    ],
)
def test_query_params_for_viewset_success_case(get_create_view_func, viewset_class):
    create_view = get_create_view_func('test', viewset_class, 'test', router=ResourceRouter())

    view = create_view('/test/', 'GET')
    view.pagination_class = None
    operation = view.schema.get_operation('/test/', 'GET')
    parameters = operation.get('parameters', [])

    assert parameters != []
